from dataclasses import dataclass
import datetime
from typing import List, Optional

from toloka.client.assignment import Assignment
from toloka.client.user_restriction import UserRestriction
from pandas import Timedelta

from .rule import (
    Rule,
    AssignmentAccuracyPredicate,
    AssignmentDurationPredicate,
    ComparisonType,
    SetAssignmentStatus,
    GiveBonusToUser,
    BlockUser,
    PredicateExpression,
    BooleanOperator,
)


def get_speed_rules(
    duration: Timedelta,
    less_or_equal: float,
    reject: bool = False,
    greater: Optional[float] = None,
) -> List[Rule]:
    rules = []
    if greater is None:
        predicate = AssignmentDurationPredicate(
            threshold=less_or_equal,
            comparison=ComparisonType.LESS_OR_EQUAL,
        )
        private_comment = f'Fast submits, <= {less_or_equal}'
    else:
        predicate = PredicateExpression(
            boolean_operator=BooleanOperator.AND,
            predicates=[
                AssignmentDurationPredicate(
                    threshold=less_or_equal,
                    comparison=ComparisonType.LESS_OR_EQUAL,
                ),
                AssignmentDurationPredicate(
                    threshold=greater,
                    comparison=ComparisonType.GREATER,
                ),
            ],
        )
        private_comment = f'Fast submits, {greater} < time <= {less_or_equal}'

    rules.append(
        Rule(
            predicate=predicate,
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment=private_comment,
                duration=duration.to_pytimedelta(),
            ),
        )
    )

    if reject:
        rules.append(
            Rule(
                predicate=predicate,
                action=SetAssignmentStatus(status=Assignment.REJECTED),
            )
        )

    return rules


@dataclass(order=True)
class BlockTimePicker:
    ratio: float = 0.05
    time_value: str = '30m'
    reject: bool = False

    @property
    def duration(self) -> Timedelta:
        return Timedelta(self.time_value)

    def __post_init__(self):
        assert 0 <= self.ratio <= 1


class RuleBuilder:
    def __init__(self):
        self.rules: List[Rule] = []
        self.reward_added: bool = False
        self.speed_control_added: bool = False
        self.control_task_control_added = False

    def add_static_reward(self, threshold: float) -> 'RuleBuilder':
        assert not self.reward_added
        assert 1.0 >= threshold >= 0.0
        self.rules.extend(
            [
                Rule(
                    predicate=AssignmentAccuracyPredicate(
                        threshold=threshold, comparison=ComparisonType.GREATER_OR_EQUAL
                    ),
                    action=SetAssignmentStatus(status=Assignment.ACCEPTED),
                ),
                Rule(
                    predicate=AssignmentAccuracyPredicate(threshold=threshold, comparison=ComparisonType.LESS),
                    action=SetAssignmentStatus(status=Assignment.REJECTED),
                ),
            ]
        )
        self.reward_added = True
        return self

    def add_dynamic_reward(
        self,
        min_bonus_amount_usd: float,
        max_bonus_amount_usd: float,
        min_accuracy_for_bonus: float = 0.0,
        min_accuracy_for_accept: float = 0.0,
        bonus_granularity_num: int = 3,
    ) -> 'RuleBuilder':
        assert not self.reward_added
        assert 0 <= min_accuracy_for_accept <= min_accuracy_for_bonus <= 1.0
        assert 0.01 <= min_bonus_amount_usd <= max_bonus_amount_usd

        if bonus_granularity_num == 1:
            assert min_accuracy_for_bonus != 0.0

        min_accuracy_for_accept_int = round(min_accuracy_for_accept * 100)
        min_accuracy_for_bonus_int = round(min_accuracy_for_bonus * 100)

        min_bonus_amount_cents = round(min_bonus_amount_usd * 100)
        max_bonus_amount_cents = round(max_bonus_amount_usd * 100)

        assert bonus_granularity_num - 1 <= max_bonus_amount_cents - min_bonus_amount_cents

        self.rules.append(
            Rule(
                predicate=AssignmentAccuracyPredicate(
                    threshold=min_accuracy_for_accept, comparison=ComparisonType.GREATER_OR_EQUAL
                ),
                action=SetAssignmentStatus(status=Assignment.ACCEPTED),
            )
        )

        if min_accuracy_for_accept_int != 0:
            self.rules.append(
                Rule(
                    predicate=AssignmentAccuracyPredicate(
                        threshold=min_accuracy_for_accept, comparison=ComparisonType.LESS
                    ),
                    action=SetAssignmentStatus(status=Assignment.REJECTED),
                )
            )

        accuracy_delta = (100 - min_accuracy_for_bonus_int) / bonus_granularity_num

        if bonus_granularity_num > 1:
            bonus_amount_delta = (max_bonus_amount_cents - min_bonus_amount_cents) / (bonus_granularity_num - 1)
        else:
            bonus_amount_delta = 0.0

        for step in range(bonus_granularity_num):
            left = (min_accuracy_for_bonus_int + step * accuracy_delta) / 100
            right = (min_accuracy_for_bonus_int + (step + 1) * accuracy_delta) / 100

            bonus_amount_usd = round(min_bonus_amount_cents + bonus_amount_delta * step) / 100

            predicates = [AssignmentAccuracyPredicate(threshold=left, comparison=ComparisonType.GREATER_OR_EQUAL)]

            # for better float comparison in edge case
            if step != bonus_granularity_num - 1:
                predicates.append(AssignmentAccuracyPredicate(threshold=right, comparison=ComparisonType.LESS))

            self.rules.append(
                Rule(
                    predicate=PredicateExpression(boolean_operator=BooleanOperator.AND, predicates=predicates),
                    action=GiveBonusToUser(amount_usd=bonus_amount_usd),
                )
            )
        self.reward_added = True
        return self

    def add_speed_control(
        self,
        ratio_rand: float,
        ratio_poor: float,
    ) -> 'RuleBuilder':
        return self.add_complex_speed_control(
            speed_blocks=[BlockTimePicker(ratio_rand, '1d', True), BlockTimePicker(ratio_poor, '8h', False)],
        )

    def add_complex_speed_control(
        self,
        speed_blocks: List[BlockTimePicker],
    ) -> 'RuleBuilder':
        assert not self.speed_control_added

        if not speed_blocks:
            return self

        speed_blocks = sorted(speed_blocks)
        block = speed_blocks[0]
        speed_rules = get_speed_rules(block.duration, block.ratio, block.reject)

        for prev, block in zip(speed_blocks, speed_blocks[1:]):
            speed_rules.extend(get_speed_rules(block.duration, block.ratio, block.reject, prev.ratio))

        self.rules.extend(speed_rules)
        self.speed_control_added = True
        return self

    def add_control_task_control(
        self,
        control_task_count: int,
        control_task_correct_count_for_hard_block: int,
        control_task_correct_count_for_soft_block: int,
    ) -> 'RuleBuilder':
        control_task_correct_ratio_for_hard_block = control_task_correct_count_for_hard_block / control_task_count
        control_task_correct_ratio_for_soft_block = control_task_correct_count_for_soft_block / control_task_count
        if control_task_correct_count_for_hard_block:
            self.rules.append(
                Rule(
                    predicate=AssignmentAccuracyPredicate(
                        threshold=control_task_correct_ratio_for_hard_block, comparison=ComparisonType.LESS
                    ),
                    action=BlockUser(
                        private_comment=f'Control tasks: [0, {control_task_correct_count_for_hard_block}) done correctly',
                        duration=datetime.timedelta(hours=8),
                        scope=UserRestriction.POOL,
                    ),
                ),  # maybe block in project
            )
        self.rules.append(
            Rule(
                predicate=PredicateExpression(
                    predicates=[
                        AssignmentAccuracyPredicate(
                            threshold=control_task_correct_ratio_for_hard_block,
                            comparison=ComparisonType.GREATER_OR_EQUAL,
                        ),
                        AssignmentAccuracyPredicate(
                            threshold=control_task_correct_ratio_for_soft_block, comparison=ComparisonType.LESS
                        ),
                    ],
                    boolean_operator=BooleanOperator.AND,
                ),
                action=BlockUser(
                    private_comment=f'Control tasks: [{control_task_correct_count_for_hard_block}, {control_task_correct_count_for_soft_block}) done correctly',
                    duration=datetime.timedelta(hours=1),
                    scope=UserRestriction.POOL,
                ),
            ),
        )

        self.control_task_control_added = True
        return self

    def build(self) -> List[Rule]:
        assert self.reward_added
        return self.rules



# Not found in contrib/python, copy-pasted from
# https://github.com/gkovacs/lz-string-python/blob/master/lzstring/__init__.py

from builtins import object
import math

keyStrBase64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/="
keyStrUriSafe = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-$"
baseReverseDic = {}


class Object(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


def getBaseValue(alphabet, character):
    if alphabet not in baseReverseDic:
        baseReverseDic[alphabet] = {}
    for i in range(len(alphabet)):
        baseReverseDic[alphabet][alphabet[i]] = i
    return baseReverseDic[alphabet][character]


def _compress(uncompressed, bitsPerChar, getCharFromInt):
    if uncompressed is None:
        return ""

    context_dictionary = {}
    context_dictionaryToCreate = {}
    context_c = ""
    context_wc = ""
    context_w = ""
    context_enlargeIn = 2  # Compensate for the first entry which should not count
    context_dictSize = 3
    context_numBits = 2
    context_data = []
    context_data_val = 0
    context_data_position = 0

    for ii in range(len(uncompressed)):
        if isinstance(uncompressed, (bytes)):
            context_c = chr(uncompressed[ii])
        else:
            context_c = uncompressed[ii]
        if context_c not in context_dictionary:
            context_dictionary[context_c] = context_dictSize
            context_dictSize += 1
            context_dictionaryToCreate[context_c] = True

        context_wc = context_w + context_c
        if context_wc in context_dictionary:
            context_w = context_wc
        else:
            if context_w in context_dictionaryToCreate:
                if ord(context_w[0]) < 256:
                    for i in range(context_numBits):
                        context_data_val = context_data_val << 1
                        if context_data_position == bitsPerChar - 1:
                            context_data_position = 0
                            context_data.append(getCharFromInt(context_data_val))
                            context_data_val = 0
                        else:
                            context_data_position += 1
                    value = ord(context_w[0])
                    for i in range(8):
                        context_data_val = (context_data_val << 1) | (value & 1)
                        if context_data_position == bitsPerChar - 1:
                            context_data_position = 0
                            context_data.append(getCharFromInt(context_data_val))
                            context_data_val = 0
                        else:
                            context_data_position += 1
                        value = value >> 1

                else:
                    value = 1
                    for i in range(context_numBits):
                        context_data_val = (context_data_val << 1) | value
                        if context_data_position == bitsPerChar - 1:
                            context_data_position = 0
                            context_data.append(getCharFromInt(context_data_val))
                            context_data_val = 0
                        else:
                            context_data_position += 1
                        value = 0
                    value = ord(context_w[0])
                    for i in range(16):
                        context_data_val = (context_data_val << 1) | (value & 1)
                        if context_data_position == bitsPerChar - 1:
                            context_data_position = 0
                            context_data.append(getCharFromInt(context_data_val))
                            context_data_val = 0
                        else:
                            context_data_position += 1
                        value = value >> 1
                context_enlargeIn -= 1
                if context_enlargeIn == 0:
                    context_enlargeIn = math.pow(2, context_numBits)
                    context_numBits += 1
                del context_dictionaryToCreate[context_w]
            else:
                value = context_dictionary[context_w]
                for i in range(context_numBits):
                    context_data_val = (context_data_val << 1) | (value & 1)
                    if context_data_position == bitsPerChar - 1:
                        context_data_position = 0
                        context_data.append(getCharFromInt(context_data_val))
                        context_data_val = 0
                    else:
                        context_data_position += 1
                    value = value >> 1

            context_enlargeIn -= 1
            if context_enlargeIn == 0:
                context_enlargeIn = math.pow(2, context_numBits)
                context_numBits += 1

            # Add wc to the dictionary.
            context_dictionary[context_wc] = context_dictSize
            context_dictSize += 1
            context_w = str(context_c)

    # Output the code for w.
    if context_w != "":
        if context_w in context_dictionaryToCreate:
            if ord(context_w[0]) < 256:
                for i in range(context_numBits):
                    context_data_val = context_data_val << 1
                    if context_data_position == bitsPerChar - 1:
                        context_data_position = 0
                        context_data.append(getCharFromInt(context_data_val))
                        context_data_val = 0
                    else:
                        context_data_position += 1
                value = ord(context_w[0])
                for i in range(8):
                    context_data_val = (context_data_val << 1) | (value & 1)
                    if context_data_position == bitsPerChar - 1:
                        context_data_position = 0
                        context_data.append(getCharFromInt(context_data_val))
                        context_data_val = 0
                    else:
                        context_data_position += 1
                    value = value >> 1
            else:
                value = 1
                for i in range(context_numBits):
                    context_data_val = (context_data_val << 1) | value
                    if context_data_position == bitsPerChar - 1:
                        context_data_position = 0
                        context_data.append(getCharFromInt(context_data_val))
                        context_data_val = 0
                    else:
                        context_data_position += 1
                    value = 0
                value = ord(context_w[0])
                for i in range(16):
                    context_data_val = (context_data_val << 1) | (value & 1)
                    if context_data_position == bitsPerChar - 1:
                        context_data_position = 0
                        context_data.append(getCharFromInt(context_data_val))
                        context_data_val = 0
                    else:
                        context_data_position += 1
                    value = value >> 1
            context_enlargeIn -= 1
            if context_enlargeIn == 0:
                context_enlargeIn = math.pow(2, context_numBits)
                context_numBits += 1
            del context_dictionaryToCreate[context_w]
        else:
            value = context_dictionary[context_w]
            for i in range(context_numBits):
                context_data_val = (context_data_val << 1) | (value & 1)
                if context_data_position == bitsPerChar - 1:
                    context_data_position = 0
                    context_data.append(getCharFromInt(context_data_val))
                    context_data_val = 0
                else:
                    context_data_position += 1
                value = value >> 1

    context_enlargeIn -= 1
    if context_enlargeIn == 0:
        context_enlargeIn = math.pow(2, context_numBits)
        context_numBits += 1

    # Mark the end of the stream
    value = 2
    for i in range(context_numBits):
        context_data_val = (context_data_val << 1) | (value & 1)
        if context_data_position == bitsPerChar - 1:
            context_data_position = 0
            context_data.append(getCharFromInt(context_data_val))
            context_data_val = 0
        else:
            context_data_position += 1
        value = value >> 1

    # Flush the last char
    while True:
        context_data_val = context_data_val << 1
        if context_data_position == bitsPerChar - 1:
            context_data.append(getCharFromInt(context_data_val))
            break
        else:
            context_data_position += 1

    return "".join(context_data)


def _decompress(length, resetValue, getNextValue):
    dictionary = {}
    enlargeIn = 4
    dictSize = 4
    numBits = 3
    entry = ""
    result = []

    data = Object(val=getNextValue(0), position=resetValue, index=1)

    for i in range(3):
        dictionary[i] = i

    bits = 0
    maxpower = math.pow(2, 2)
    power = 1

    while power != maxpower:
        resb = data.val & data.position
        data.position >>= 1
        if data.position == 0:
            data.position = resetValue
            data.val = getNextValue(data.index)
            data.index += 1

        bits |= power if resb > 0 else 0
        power <<= 1

    next = bits
    if next == 0:
        bits = 0
        maxpower = math.pow(2, 8)
        power = 1
        while power != maxpower:
            resb = data.val & data.position
            data.position >>= 1
            if data.position == 0:
                data.position = resetValue
                data.val = getNextValue(data.index)
                data.index += 1
            bits |= power if resb > 0 else 0
            power <<= 1
        c = chr(bits)
    elif next == 1:
        bits = 0
        maxpower = math.pow(2, 16)
        power = 1
        while power != maxpower:
            resb = data.val & data.position
            data.position >>= 1
            if data.position == 0:
                data.position = resetValue
                data.val = getNextValue(data.index)
                data.index += 1
            bits |= power if resb > 0 else 0
            power <<= 1
        c = chr(bits)
    elif next == 2:
        return ""

    dictionary[3] = c
    w = c
    result.append(c)
    counter = 0
    while True:
        counter += 1
        if data.index > length:
            return ""

        bits = 0
        maxpower = math.pow(2, numBits)
        power = 1
        while power != maxpower:
            resb = data.val & data.position
            data.position >>= 1
            if data.position == 0:
                data.position = resetValue
                data.val = getNextValue(data.index)
                data.index += 1
            bits |= power if resb > 0 else 0
            power <<= 1

        c = bits
        if c == 0:
            bits = 0
            maxpower = math.pow(2, 8)
            power = 1
            while power != maxpower:
                resb = data.val & data.position
                data.position >>= 1
                if data.position == 0:
                    data.position = resetValue
                    data.val = getNextValue(data.index)
                    data.index += 1
                bits |= power if resb > 0 else 0
                power <<= 1

            dictionary[dictSize] = chr(bits)
            dictSize += 1
            c = dictSize - 1
            enlargeIn -= 1
        elif c == 1:
            bits = 0
            maxpower = math.pow(2, 16)
            power = 1
            while power != maxpower:
                resb = data.val & data.position
                data.position >>= 1
                if data.position == 0:
                    data.position = resetValue
                    data.val = getNextValue(data.index)
                    data.index += 1
                bits |= power if resb > 0 else 0
                power <<= 1
            dictionary[dictSize] = chr(bits)
            dictSize += 1
            c = dictSize - 1
            enlargeIn -= 1
        elif c == 2:
            return "".join(result)

        if enlargeIn == 0:
            enlargeIn = math.pow(2, numBits)
            numBits += 1

        if c in dictionary:
            entry = dictionary[c]
        else:
            if c == dictSize:
                entry = w + w[0]
            else:
                return None
        result.append(entry)

        # Add w+entry[0] to the dictionary.
        dictionary[dictSize] = w + entry[0]
        dictSize += 1
        enlargeIn -= 1

        w = entry
        if enlargeIn == 0:
            enlargeIn = math.pow(2, numBits)
            numBits += 1


class LZString(object):
    @staticmethod
    def compress(uncompressed):
        return _compress(uncompressed, 16, chr)

    @staticmethod
    def compressToUint8Array(uncompressed):
        return bytes([ord(x) for x in _compress(uncompressed, 8, chr)])

    @staticmethod
    def compressToUTF16(uncompressed):
        if uncompressed is None:
            return ""
        return _compress(uncompressed, 15, lambda a: chr(a + 32)) + " "

    @staticmethod
    def compressToBase64(uncompressed):
        if uncompressed is None:
            return ""
        res = _compress(uncompressed, 6, lambda a: keyStrBase64[a])
        # To produce valid Base64
        end = len(res) % 4
        if end > 0:
            res += "=" * (4 - end)
        return res

    @staticmethod
    def compressToEncodedURIComponent(uncompressed):
        if uncompressed is None:
            return ""
        return _compress(uncompressed, 6, lambda a: keyStrUriSafe[a])

    @staticmethod
    def decompress(compressed):
        if compressed is None:
            return ""
        if compressed == "":
            return None
        return _decompress(len(compressed), 32768, lambda index: ord(compressed[index]))

    @staticmethod
    def decompressFromUint8Array(compressed):
        if compressed is None:
            return ""
        if compressed == "":
            return None
        return _decompress(len(compressed), 128, lambda index: compressed[index])

    @staticmethod
    def decompressFromUTF16(compressed):
        if compressed is None:
            return ""
        if compressed == "":
            return None
        return _decompress(len(compressed), 16384, lambda index: compressed[index] - 32)

    @staticmethod
    def decompressFromBase64(compressed):
        if compressed is None:
            return ""
        if compressed == "":
            return None
        return _decompress(len(compressed), 32, lambda index: getBaseValue(keyStrBase64, compressed[index]))

    @staticmethod
    def decompressFromEncodedURIComponent(compressed):
        if compressed is None:
            return ""
        if compressed == "":
            return None
        compressed = compressed.replace(" ", "+")
        return _decompress(len(compressed), 32, lambda index: getBaseValue(keyStrUriSafe, compressed[index]))



from datetime import timedelta
import logging
from typing import List, Optional, Union

import toloka.client as toloka

from .. import mapping, objects, pricing

logger = logging.getLogger(__name__)

min_duration_ratio = 0.4
max_duration_ratio = 2.5


def training_is_suitable(training: toloka.Training, task_duration_hint: timedelta) -> bool:
    if training.status != toloka.Training.Status.OPEN:
        # Toloka sometimes closes trainings on its own, this may be due to too many completed assignments.
        # In this case, we need to create the training again, regardless of whether it is suitable or not.
        return False
    task_duration_hint_with_overhead = task_duration_hint * pricing.training_overhead_multiplier
    existing_task_duration_hint_with_overhead = timedelta(
        seconds=(training.assignment_max_duration_seconds / training.training_tasks_in_task_suite_count)
    )
    return (
        min_duration_ratio
        <= existing_task_duration_hint_with_overhead / task_duration_hint_with_overhead
        <= max_duration_ratio
    )


def find_training(
    project_id: str,
    client: toloka.TolokaClient,
) -> Optional[toloka.Training]:
    # Toloka API doesn't support multiple training statuses in search for some reason. So we make two calls.
    found_trainings = []
    for status in (toloka.Training.Status.OPEN, toloka.Training.Status.CLOSED):
        found_trainings += client.find_trainings(
            project_id=project_id, status=status, sort=toloka.search_requests.TrainingSortItems(['-created']), limit=1
        ).items
    if not found_trainings:
        return None
    return sorted(found_trainings, key=lambda t: t.created, reverse=True)[0]  # current training


def create_training(
    project_id: str,
    private_name: str,
    training_config: pricing.TrainingConfig,
    client: toloka.TolokaClient,
    reuse_tasks_from: Optional[toloka.Training] = None,
) -> toloka.Training:
    training = toloka.training.Training(
        project_id=project_id,
        private_name=private_name,
        may_contain_adult_content=True,
        assignment_max_duration_seconds=int(training_config.max_assignment_duration.total_seconds()),
        mix_tasks_in_creation_order=True,
        shuffle_tasks_in_task_suite=True,
        training_tasks_in_task_suite_count=training_config.training_tasks_in_task_suite_count,
        task_suites_required_to_pass=training_config.task_suites_required_to_pass,
        retry_training_after_days=30,  # todo
        inherited_instructions=True,
    )
    created_training = client.create_training(training)

    if reuse_tasks_from is not None:
        existing_tasks = list(client.get_tasks(pool_id=reuse_tasks_from.id))
        for task in existing_tasks:
            task.pool_id = created_training.id
        client.create_tasks(existing_tasks, allow_defaults=True, async_mode=True, skip_invalid_items=False)

    return created_training


def _get_training_task(task_mapping: mapping.TaskMapping, solution: mapping.Objects) -> toloka.Task.KnownSolution:
    return toloka.Task.KnownSolution(output_values=task_mapping.toloka_values(solution, output=True))


def _get_training_tasks(
    task_mapping: mapping.TaskMapping, solution: Union[mapping.Objects, List[mapping.Objects]]
) -> List[toloka.Task.KnownSolution]:
    if isinstance(solution, list):
        return [_get_training_task(task_mapping, single_solution) for single_solution in solution]
    return [_get_training_task(task_mapping, solution)]


def add_training_tasks(
    training_id: str,
    task_mapping: mapping.TaskMapping,
    solutions: List[Union[mapping.TaskSingleSolution, mapping.TaskMultipleSolutionOptions]],
    comments: List[objects.Text],
    client: toloka.TolokaClient,
):
    training_tasks = [
        toloka.Task(
            pool_id=training_id,
            input_values=task_mapping.toloka_values(task),
            known_solutions=_get_training_tasks(task_mapping, solution),
            message_on_unknown_solution=comment.text,
        )
        for (task, solution), comment in zip(solutions, comments)
    ]
    logger.debug(f'Adding {len(training_tasks)} training tasks to training {training_id}')
    client.create_tasks(training_tasks, allow_defaults=True, async_mode=True, skip_invalid_items=False)


def get_skill_id_from_training(training: toloka.Pool) -> str:
    for confing in training.quality_control.configs:
        for rule in confing.rules:
            if isinstance(rule.action, toloka.actions.SetSkillFromOutputField):
                return rule.action.parameters.skill_id
    assert False, 'unreachable'



from collections import deque
from dataclasses import dataclass, field
from datetime import timedelta
from typing import List, Dict, Deque


import toloka.client as toloka

from .. import client, classification_loop, control, evaluation, pricing, worker
from ..pricing import _get_num_classes

from . import params
from .base import Event, Update, UIEvent, Node, ParamName
from .characteristics import _get_stats_and_chars, _get_annotation_stats_and_chars
from .params import Param, Characteristics, Stat
from .util import get_overlap, get_agg_algorithm


@dataclass
class EventLoop:
    event_to_subscribers: Dict[str, Node]
    params: List[Node]
    stats: List[Stat]
    characteristics: Characteristics
    state: Dict[ParamName, Param] = field(init=False)
    queue: Deque[Event] = deque()
    ui_initialized: bool = False

    def __post_init__(self):
        self.state = {p.name: p for p in self.params if isinstance(p, Param)}
        self.update_stats_and_chars()

    def update_stats_and_chars(self):
        stats, characteristics = _get_stats_and_chars(
            num_classes=self.state[ParamName.CONTROL_TASKS_FOR_ACCEPT].num_classes,
            **{
                name.value: self.state[name].value
                for name in [
                    ParamName.TASK_DURATION_HINT,
                    ParamName.TASK_COUNT,
                    ParamName.CONTROL_TASK_COUNT,
                    ParamName.ASSIGNMENT_PRICE,
                    ParamName.OVERLAP,
                    ParamName.OVERLAP_CONFIDENCE,
                    ParamName.CONTROL_TASKS_FOR_ACCEPT,
                    ParamName.VERIFIED_LANG,
                    ParamName.TRAINING,
                    ParamName.TRAINING_SCORE,
                ]
            },
        )
        for v, s in zip(stats, self.stats):
            s.process(Event(ParamName.STATISTICS.value, {s.name: Update(v)}))

        self.characteristics.process(
            Event(ParamName.CHARACTERISTICS.value, {self.characteristics.name: Update(characteristics)})
        )

    def start(self, event: UIEvent):
        if not self.ui_initialized:
            return

        changed = False
        if event.sender == ParamName.SPEED_CONTROL.ui:
            changed = self.state[ParamName.SPEED_CONTROL].check_update(event)
        else:
            for name, update in event.update.items():
                p = self.state[name]
                old = Update(p.value, p.disabled)
                if old == update:
                    continue
                changed = True
                break
        if not changed:
            # redundant event from dispatch
            return

        self.queue.append(event)

        while self.queue:
            e = self.queue.popleft()
            for p in self.event_to_subscribers.get(e.sender, []):
                self.queue.extend(p.process(e))

        self.update_stats_and_chars()

    def get_params(self) -> client.Params:
        # todo: proxy age_restriction to pool builder
        task_duration_hint = self.state[ParamName.TASK_DURATION_HINT].value
        control_task_count = self.state[ParamName.CONTROL_TASK_COUNT].value

        correct_control_task_ratio_for_acceptance = (
            0.0
            if control_task_count == 0
            else self.state[ParamName.CONTROL_TASKS_FOR_ACCEPT].value / control_task_count
        )
        builder = (
            control.RuleBuilder()
            .add_static_reward(threshold=correct_control_task_ratio_for_acceptance)
            .add_complex_speed_control(self.state[ParamName.SPEED_CONTROL].value)
        )

        if control_task_count > 0:
            builder = builder.add_control_task_control(
                control_task_count=control_task_count,
                control_task_correct_count_for_hard_block=0,
                control_task_correct_count_for_soft_block=self.state[ParamName.CONTROL_TASKS_FOR_BLOCK].value,
            )

        training_score = None
        if self.state[ParamName.TRAINING].value:
            training_score = self.state[ParamName.TRAINING_SCORE].value
        return client.Params(
            worker_filter=worker.WorkerFilter(
                filters=[
                    worker.WorkerFilter.Params(
                        lang=worker.LanguageRequirement(
                            lang=self.state[ParamName.LANGUAGE].value,
                            verified=self.state[ParamName.VERIFIED_LANG].value,
                        ),
                        regions=worker.lang_to_default_regions.get(self.state[ParamName.LANGUAGE].value, {}),
                    )
                ],
                training_score=training_score,
            ),
            task_duration_hint=task_duration_hint,
            pricing_config=pricing.PoolPricingConfig(
                real_tasks_count=self.state[ParamName.TASK_COUNT].value,
                control_tasks_count=control_task_count,
                assignment_price=self.state[ParamName.ASSIGNMENT_PRICE].value,
            ),
            overlap=get_overlap(self.state[ParamName.OVERLAP].value, self.state[ParamName.OVERLAP_CONFIDENCE].value),
            aggregation_algorithm=get_agg_algorithm(self.state[ParamName.AGGREGATION_ALGORITHM].value),
            control=control.Control(
                rules=builder.build(),
            ),
            task_duration_function=None,
        )


class AnnotationEventLoop(EventLoop):
    def update_stats_and_chars(self):
        stats, characteristics = _get_annotation_stats_and_chars(
            **{
                name.value: self.state[name].value
                for name in [
                    ParamName.TASK_DURATION_HINT,
                    ParamName.TASK_COUNT,
                    ParamName.ASSIGNMENT_PRICE,
                    ParamName.CHECK_SAMPLE,
                    ParamName.MAX_TASKS_TO_CHECK,
                    ParamName.ACCURACY_THRESHOLD,
                    ParamName.OVERLAP,
                    ParamName.OVERLAP_CONFIDENCE,
                    ParamName.CONTROL_TASKS_FOR_ACCEPT,
                    ParamName.VERIFIED_LANG,
                ]
            },
        )
        for v, s in zip(stats, self.stats):
            s.process(Event(ParamName.STATISTICS.value, {s.name: Update(v)}))

        self.characteristics.process(
            Event(ParamName.CHARACTERISTICS.value, {self.characteristics.name: Update(characteristics)})
        )

    def get_params(self) -> client.AnnotationParams:
        # todo: age_restriction to pool builder
        task_duration_hint = self.state[ParamName.TASK_DURATION_HINT].value
        checked_task_count = self.state[ParamName.TASK_COUNT].value
        assignment_check_sample = None

        if self.state[ParamName.CHECK_SAMPLE].value:
            checked_task_count = self.state[ParamName.MAX_TASKS_TO_CHECK].value
            assignment_accuracy_finalization_threshold = (
                0.0 if checked_task_count == 0 else self.state[ParamName.ACCURACY_THRESHOLD].value / checked_task_count
            )
            assignment_check_sample = evaluation.AssignmentCheckSample(
                max_tasks_to_check=checked_task_count,
                assignment_accuracy_finalization_threshold=assignment_accuracy_finalization_threshold,
            )
        correct_checked_task_ratio_for_acceptance = (
            0.0
            if checked_task_count == 0
            else self.state[ParamName.CONTROL_TASKS_FOR_ACCEPT].value / checked_task_count
        )
        builder = (
            control.RuleBuilder()
            .add_static_reward(threshold=correct_checked_task_ratio_for_acceptance)
            .add_complex_speed_control(self.state[ParamName.SPEED_CONTROL].value)
        )

        training_score = None
        return client.AnnotationParams(
            worker_filter=worker.WorkerFilter(
                filters=[
                    worker.WorkerFilter.Params(
                        lang=worker.LanguageRequirement(
                            lang=self.state[ParamName.LANGUAGE].value,
                            verified=self.state[ParamName.VERIFIED_LANG].value,
                        ),
                        regions=worker.lang_to_default_regions.get(self.state[ParamName.LANGUAGE].value, {}),
                    )
                ],
                training_score=training_score,
            ),
            task_duration_hint=task_duration_hint,
            pricing_config=pricing.PoolPricingConfig(
                real_tasks_count=self.state[ParamName.TASK_COUNT].value,
                control_tasks_count=0,
                assignment_price=self.state[ParamName.ASSIGNMENT_PRICE].value,
            ),
            overlap=classification_loop.DynamicOverlap(
                min_overlap=1,
                max_overlap=int(self.state[ParamName.OVERLAP].value),
                confidence=self.state[ParamName.OVERLAP_CONFIDENCE].value,
            ),
            aggregation_algorithm=None,
            control=control.Control(
                rules=builder.build(),
            ),
            assignment_check_sample=assignment_check_sample,
            task_duration_function=None,
        )


def get_loop(
    task_spec: client.PreparedTaskSpec,
    task_duration_hint: timedelta,
    toloka_client: toloka.TolokaClient,
) -> EventLoop:
    quality = params.Stat(ParamName.QUALITY, 'Quality')
    time_ = params.Stat(ParamName.TIME, 'Time', reverse=True)
    cost = params.Stat(ParamName.COST, 'Cost', reverse=True)

    help_p = params.Help()
    task_duration_p = params.TaskDuration(task_duration_hint)
    lang_p = params.Lang(task_spec)
    training_p = params.Training(task_spec, toloka_client)
    task_count_p = params.TaskCount(task_duration_hint)
    quality_preset_p = params.QualityPreset()
    control_task_count_p = params.ControlTaskCount(task_count_p.value, quality_preset_p.value)
    assignment_price_p = params.AssignmentPrice(task_duration_hint, task_count_p.value, control_task_count_p.value)
    age_restriction_p = params.AgeRestriction()
    verified_lang_p = params.VerifiedLang(task_spec)
    training_value_p = params.TrainingValue(training_p.value)
    overlap_p = params.Overlap(quality_preset_p.value)
    overlap_confidence_p = params.OverlapConfidence(quality_preset_p.value, overlap_p.value)
    agg_p = params.AggregationAlgorithm()
    control_tasks_for_accept_p = params.ControlTasksForAccept(
        _get_num_classes(task_spec.task_mapping),
        quality_preset_p.value,
        control_task_count_p.value,
    )
    control_tasks_for_block_p = params.ControlTasksForBlock(
        quality_preset_p.value,
        control_task_count_p.value,
    )
    speed_control_p = params.SpeedControlList(quality_preset_p.value)

    general_quality_group = params.ParamGroup(
        name=ParamName.GENERAL_QUALITY_GROUP,
        params=[task_duration_p, lang_p, quality_preset_p],
    )

    quality_tasks_group = params.ParamGroup(
        name=ParamName.QUALITY_TASKS_GROUP,
        params=[task_duration_p, quality_preset_p, task_count_p],
    )

    quality_assignment_group = params.ParamGroupControlTasksRecalculate(
        name=ParamName.QUALITY_ASSIGNMENT_GROUP,
        params=[task_duration_p, quality_preset_p, task_count_p, control_task_count_p],
    )
    created_params = [
        help_p,
        task_duration_p,
        lang_p,
        task_count_p,
        control_task_count_p,
        control_tasks_for_accept_p,
        control_tasks_for_block_p,
        quality_preset_p,
        overlap_p,
        overlap_confidence_p,
        agg_p,
        assignment_price_p,
        age_restriction_p,
        verified_lang_p,
        training_p,
        training_value_p,
        speed_control_p,
        general_quality_group,
        quality_tasks_group,
        quality_assignment_group,
    ]
    stats = [quality, time_, cost]
    characteristics = params.Characteristics()
    return EventLoop(
        event_to_subscribers={
            ParamName.QUALITY_PRESET.ui: [general_quality_group],
            ParamName.GENERAL_QUALITY_GROUP.value: [quality_tasks_group, overlap_p, speed_control_p],
            ParamName.TASK_COUNT.ui: [quality_tasks_group],
            ParamName.QUALITY_TASKS_GROUP.value: [quality_assignment_group],
            ParamName.CONTROL_TASK_COUNT.ui: [quality_assignment_group],
            ParamName.QUALITY_ASSIGNMENT_GROUP.value: [
                task_count_p,
                control_task_count_p,
                assignment_price_p,
                quality_preset_p,
                control_tasks_for_accept_p,
                control_tasks_for_block_p,
            ],
            ParamName.OVERLAP.ui: [overlap_p],
            ParamName.OVERLAP.value: [overlap_confidence_p],
            ParamName.OVERLAP_CONFIDENCE.ui: [overlap_confidence_p],
            ParamName.ASSIGNMENT_PRICE.ui: [assignment_price_p],
            ParamName.CONTROL_TASKS_FOR_ACCEPT.ui: [control_tasks_for_accept_p],
            ParamName.CONTROL_TASKS_FOR_BLOCK.ui: [control_tasks_for_block_p],
            ParamName.AGGREGATION_ALGORITHM.ui: [agg_p],
            ParamName.TRAINING.ui: [training_p],
            ParamName.TRAINING.value: [training_value_p],
            ParamName.TRAINING_SCORE.ui: [training_value_p],
            ParamName.AGE_RESTRICTION.ui: [age_restriction_p],
            ParamName.VERIFIED_LANG.ui: [verified_lang_p],
            ParamName.HELP.value: [help_p],
            ParamName.SPEED_CONTROL.ui: [speed_control_p],
        },
        params=created_params,
        stats=stats,
        characteristics=characteristics,
    )


def get_loop_for_annotation(
    task_spec: client.PreparedTaskSpec,
    task_duration_hint: timedelta,
    toloka_client: toloka.TolokaClient,
) -> AnnotationEventLoop:
    quality = params.Stat(ParamName.QUALITY, 'Quality')
    time_ = params.Stat(ParamName.TIME, 'Time', reverse=True)
    cost = params.Stat(ParamName.COST, 'Cost', reverse=True)

    help_p = params.Help()
    task_duration_p = params.TaskDuration(task_duration_hint)
    lang_p = params.Lang(task_spec)
    task_count_p = params.TaskCount(task_duration_hint)
    quality_preset_p = params.AnnotationQualityPreset()
    assignment_price_p = params.AnnotationAssignmentPrice(task_duration_hint, task_count_p.value)
    age_restriction_p = params.AgeRestriction()
    verified_lang_p = params.VerifiedLang(task_spec)
    max_attempts_p = params.MaxAttempts(quality_preset_p.value)
    check_sample_p = params.CheckSample(quality_preset_p.value)
    max_tasks_to_check_p = params.MaxTasksToCheck(quality_preset_p.value, task_count_p.value, check_sample_p.value)
    accuracy_threshold_p = params.AccuracyThreshold(
        quality_preset_p.value, task_count_p.value, check_sample_p.value, max_tasks_to_check_p.value
    )
    overlap_confidence_p = params.Confidence(quality_preset_p.value, max_attempts_p.value)
    control_tasks_for_accept_p = params.CheckedTasksForAccept(
        quality_preset_p.value,
        task_count_p.value,
    )
    control_tasks_for_block_p = params.CheckedTasksForBlock(
        quality_preset_p.value,
        task_count_p.value,
    )
    speed_control_p = params.SpeedControlList(quality_preset_p.value)

    general_quality_group = params.ParamGroup(
        name=ParamName.GENERAL_QUALITY_GROUP,
        params=[task_duration_p, lang_p, quality_preset_p],
    )

    quality_tasks_group = params.ParamGroup(
        name=ParamName.QUALITY_TASKS_GROUP,
        params=[task_duration_p, quality_preset_p, task_count_p],
    )

    quality_assignment_group = params.ParamGroup(
        name=ParamName.QUALITY_ASSIGNMENT_GROUP,
        params=[task_duration_p, quality_preset_p, task_count_p, check_sample_p],
    )

    check_sample_group = params.ParamGroupCheckedTasksRecalculate(
        name=ParamName.CHECK_SAMPLE_GROUP,
        params=[quality_preset_p, task_count_p, check_sample_p, max_tasks_to_check_p],
    )
    created_params = [
        help_p,
        task_duration_p,
        lang_p,
        task_count_p,
        check_sample_p,
        max_tasks_to_check_p,
        accuracy_threshold_p,
        control_tasks_for_accept_p,
        control_tasks_for_block_p,
        quality_preset_p,
        max_attempts_p,
        overlap_confidence_p,
        assignment_price_p,
        age_restriction_p,
        verified_lang_p,
        speed_control_p,
        general_quality_group,
        quality_tasks_group,
        quality_assignment_group,
        check_sample_group,
    ]
    stats = [quality, time_, cost]
    characteristics = params.Characteristics()
    return AnnotationEventLoop(
        event_to_subscribers={
            ParamName.QUALITY_PRESET.ui: [general_quality_group],
            ParamName.GENERAL_QUALITY_GROUP.value: [quality_tasks_group, max_attempts_p, speed_control_p],
            ParamName.TASK_COUNT.ui: [quality_tasks_group],
            ParamName.QUALITY_TASKS_GROUP.value: [quality_assignment_group],
            ParamName.CHECK_SAMPLE.ui: [quality_assignment_group],
            ParamName.MAX_TASKS_TO_CHECK.ui: [check_sample_group],
            ParamName.QUALITY_ASSIGNMENT_GROUP.value: [
                check_sample_group,
                assignment_price_p,
            ],
            ParamName.CHECK_SAMPLE_GROUP.value: [
                task_count_p,
                check_sample_p,
                quality_preset_p,
                max_tasks_to_check_p,
                accuracy_threshold_p,
                control_tasks_for_accept_p,
                control_tasks_for_block_p,
            ],
            ParamName.OVERLAP.ui: [max_attempts_p],
            ParamName.OVERLAP.value: [overlap_confidence_p],
            ParamName.OVERLAP_CONFIDENCE.ui: [overlap_confidence_p],
            ParamName.ASSIGNMENT_PRICE.ui: [assignment_price_p],
            ParamName.CONTROL_TASKS_FOR_ACCEPT.ui: [control_tasks_for_accept_p],
            ParamName.CONTROL_TASKS_FOR_BLOCK.ui: [control_tasks_for_block_p],
            ParamName.ACCURACY_THRESHOLD.ui: [accuracy_threshold_p],
            ParamName.AGE_RESTRICTION.ui: [age_restriction_p],
            ParamName.VERIFIED_LANG.ui: [verified_lang_p],
            ParamName.HELP.value: [help_p],
            ParamName.SPEED_CONTROL.ui: [speed_control_p],
        },
        params=created_params,
        stats=stats,
        characteristics=characteristics,
    )



import abc
from dataclasses import dataclass
import datetime
from decimal import Decimal
import enum
from typing import List, Union, Optional, Dict, Any


import toloka.client as toloka
from toloka.client import TolokaClient
from toloka.client.user_restriction import (
    UserRestriction,
    PoolUserRestriction,
    ProjectUserRestriction,
    AllProjectsUserRestriction,
)
from toloka.client.assignment import Assignment, AssignmentPatch
from toloka.client.user_bonus import UserBonus


class Predicate:
    @abc.abstractmethod
    def check(self, **kwargs) -> bool:
        ...


class ComparisonType(enum.Enum):
    GREATER_OR_EQUAL = '>='
    GREATER = '>'
    LESS_OR_EQUAL = '<='
    LESS = '<'


PredicateValue = Union[float, datetime.timedelta]


@dataclass
class ThresholdComparisonPredicate(Predicate):
    threshold: PredicateValue
    comparison: ComparisonType

    def check(self, value: PredicateValue, **kwargs) -> bool:
        threshold = self.get_threshold(**kwargs)
        return {
            ComparisonType.GREATER_OR_EQUAL: lambda: value >= threshold,
            ComparisonType.GREATER: lambda: value > threshold,
            ComparisonType.LESS_OR_EQUAL: lambda: value <= threshold,
            ComparisonType.LESS: lambda: value < threshold,
        }[self.comparison]()

    def get_threshold(self, **kwargs) -> PredicateValue:
        return self.threshold


@dataclass
class AssignmentAccuracyPredicate(ThresholdComparisonPredicate):
    ...


@dataclass
class AssignmentDurationPredicate(ThresholdComparisonPredicate):
    def get_threshold(self, assignment_duration_hint: datetime.timedelta, **kwargs) -> datetime.timedelta:
        return self.threshold * assignment_duration_hint


@dataclass
class AlwaysTruePredicate(Predicate):
    def check(self, **kwargs) -> bool:
        return True


class BooleanOperator(enum.Enum):
    AND = 'and'
    OR = 'or'


@dataclass
class PredicateExpression(Predicate):
    boolean_operator: BooleanOperator
    predicates: List[Predicate]

    def __post_init__(self):
        assert self.predicates, 'List of predicates should be non-empty'
        predicate_type = type(self.predicates[0])
        assert not issubclass(predicate_type, PredicateExpression), 'Nested expressions are not allowed'
        for predicate in self.predicates:
            assert isinstance(predicate, predicate_type), (
                'All predicates should be of same type. ' f'Expected {predicate_type}, got {type(predicate)}'
            )

    def check(self, **kwargs) -> bool:
        if self.boolean_operator == BooleanOperator.AND:
            return all(predicate.check(**kwargs) for predicate in self.predicates)
        elif self.boolean_operator == BooleanOperator.OR:
            return any(predicate.check(**kwargs) for predicate in self.predicates)
        else:
            raise ValueError(f'unsupported boolean operator for predicates: {self.boolean_operator}')


class Action:
    @abc.abstractmethod
    def perform(self, **kwargs):
        ...


@dataclass
class BlockUser(Action):
    scope: UserRestriction.Scope
    private_comment: str
    duration: Optional[datetime.timedelta] = None

    def perform(
        self,
        client: TolokaClient,
        user_id: str,
        assignment_status: toloka.Assignment.Status,
        block_start: Optional[datetime.datetime] = None,
        pool_id: Optional[str] = None,
        project_id: Optional[str] = None,
        **kwargs,
    ) -> UserRestriction:
        params = {'user_id': user_id, 'private_comment': self.private_comment}
        user_restriction: UserRestriction
        if self.duration:
            start = block_start or datetime.datetime.now()
            will_expire = start + self.duration
            params['will_expire'] = will_expire
        if self.scope == UserRestriction.Scope.POOL:
            user_restriction = PoolUserRestriction(pool_id=pool_id, **params)
        elif self.scope == UserRestriction.Scope.PROJECT:
            user_restriction = ProjectUserRestriction(project_id=project_id, **params)
        elif self.scope == UserRestriction.Scope.ALL:
            user_restriction = AllProjectsUserRestriction(**params)
        else:
            raise ValueError(f'unsupported scope for block: {self.scope}')
        if (
            assignment_status != toloka.Assignment.Status.ACCEPTED
            and assignment_status != toloka.Assignment.Status.REJECTED
        ):
            client.set_user_restriction(user_restriction=user_restriction)
        return user_restriction


@dataclass
class GiveBonusToUser(Action):
    amount_usd: float

    def perform(self, user_id: str, assignment_id: str, **kwargs) -> UserBonus:
        return UserBonus(
            user_id=user_id, amount=Decimal(self.amount_usd), assignment_id=assignment_id, without_message=True
        )  # we may want to send the message here


@dataclass
class SetAssignmentStatus(Action):
    status: Assignment.Status

    def perform(
        self,
        client: TolokaClient,
        assignment: toloka.Assignment,
        public_comment: str,
        **kwargs,
    ) -> Assignment.Status:
        if not assignment.id:
            # model workers have stub assignment
            return self.status
        if (
            assignment.status == toloka.Assignment.Status.REJECTED
            or assignment.status == toloka.Assignment.Status.ACCEPTED
        ):
            return assignment.status
        comment = public_comment if self.status == Assignment.REJECTED else ''
        client.patch_assignment(
            assignment_id=assignment.id, patch=AssignmentPatch(public_comment=comment, status=self.status)
        )
        return self.status


@dataclass
class Rule:
    predicate: Predicate
    action: Action

    def apply(self, **kwargs):
        if self.predicate.check(**kwargs):
            return self.action.perform(**kwargs)


# List of all rules for pool
@dataclass
class Control:
    rules: List[Rule]

    def filter_rules(self, predicate_type: type, action_type: type) -> List[Rule]:
        assert issubclass(predicate_type, Predicate)
        assert issubclass(action_type, Action)
        assert not issubclass(predicate_type, PredicateExpression), 'You should filter by inner predicate type'

        def is_suitable(predicate: Predicate) -> bool:
            if not isinstance(predicate, PredicateExpression):
                return isinstance(predicate, predicate_type)
            return all(isinstance(inner, predicate_type) for inner in predicate.predicates)

        return [rule for rule in self.rules if is_suitable(rule.predicate) and isinstance(rule.action, action_type)]

    def to_toloka_quality_control(
        self, control_tasks_count: int, assignment_duration_hint: datetime.timedelta
    ) -> List[Dict[str, Any]]:
        converted_rules = []
        for rule in self.filter_rules(AssignmentDurationPredicate, BlockUser):
            threshold = None
            fast_count, total_count = 1, 1

            if isinstance(rule.predicate, AssignmentDurationPredicate):
                threshold = rule.predicate.threshold
            elif isinstance(rule.predicate, PredicateExpression):
                for predicate in rule.predicate.predicates:
                    assert isinstance(predicate, AssignmentDurationPredicate)
                    if predicate.comparison == ComparisonType('<=') or predicate.comparison == ComparisonType('<'):
                        threshold = predicate.threshold
            else:
                assert False, 'unreachable'

            assert threshold is not None
            assert assignment_duration_hint is not None
            assert isinstance(rule.action, BlockUser)

            assignment_threshold = assignment_duration_hint * threshold

            converted_rules.append(
                dict(
                    collector=toloka.collectors.AssignmentSubmitTime(
                        fast_submit_threshold_seconds=max(int(assignment_threshold.total_seconds()), 1),
                        history_size=10,
                    ),
                    conditions=[
                        toloka.conditions.TotalSubmittedCount >= total_count,
                        toloka.conditions.FastSubmittedCount >= fast_count,
                    ],
                    action=toloka.actions.RestrictionV2(
                        private_comment=rule.action.private_comment,
                        duration=int(rule.action.duration.total_seconds() // 60),
                        duration_unit=toloka.user_restriction.DurationUnit.MINUTES,
                        scope=toloka.UserRestriction.PROJECT,
                    ),
                )
            )

        for rule in self.filter_rules(AssignmentAccuracyPredicate, BlockUser):
            assert isinstance(rule.action, BlockUser)
            condition_assignment_answers = toloka.conditions.TotalAnswersCount >= control_tasks_count

            def convert_predicate_to_condition(accuracy_predicate: AssignmentAccuracyPredicate):
                value = accuracy_predicate.threshold * 100
                if accuracy_predicate.comparison == ComparisonType.LESS:
                    return toloka.conditions.CorrectAnswersRate < value
                elif accuracy_predicate.comparison == ComparisonType.LESS_OR_EQUAL:
                    return toloka.conditions.CorrectAnswersRate <= value
                elif accuracy_predicate.comparison == ComparisonType.GREATER:
                    return toloka.conditions.CorrectAnswersRate > value
                elif accuracy_predicate.comparison == ComparisonType.GREATER_OR_EQUAL:
                    return toloka.conditions.CorrectAnswersRate >= value
                else:
                    raise ValueError(f'unknown comparison type: {accuracy_predicate.comparison}')

            if isinstance(rule.predicate, PredicateExpression):
                predicate_conditions = [
                    convert_predicate_to_condition(predicate) for predicate in rule.predicate.predicates
                ]
            elif isinstance(rule.predicate, AssignmentAccuracyPredicate):
                predicate_conditions = [convert_predicate_to_condition(rule.predicate)]

            conditions = [condition_assignment_answers] + predicate_conditions

            converted_rules.append(
                dict(
                    collector=toloka.collectors.GoldenSet(history_size=control_tasks_count),
                    conditions=conditions,
                    action=toloka.actions.RestrictionV2(
                        private_comment=rule.action.private_comment,
                        duration=int(rule.action.duration.total_seconds() // 60),
                        duration_unit=toloka.user_restriction.DurationUnit.MINUTES,
                        scope=toloka.UserRestriction.PROJECT,
                    ),
                )
            )
        return converted_rules



from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
import logging
from multiprocessing.pool import ThreadPool
from random import shuffle
from time import sleep
from typing import List, Dict, Type, Optional, Tuple, Any, Callable

import numpy as np
from toloka.client.exceptions import IncorrectActionsApiError, DoesNotExistApiError

from .client import NDATolokaClient, NDARegisterClient
from .util import WorkerSpeed, TaskInfo, StoppableThread, content, swap_task, mutate

from dataforge import base, client as df_client, mapping

logger = logging.getLogger(__name__)

true, false = base.BinaryEvaluation(True), base.BinaryEvaluation(False)


def softmax(x: np.array) -> np.array:
    e = np.exp(x)
    s = np.sum(e)
    return e / s


@dataclass
class WorkerConfig:
    speed: WorkerSpeed
    # качественность работника
    quality: float

    obj_cls: Type[base.Class]
    confusion_matrix: Dict[base.Class, Dict[base.Class, float]] = field(init=False)
    random_vector: Dict[base.Class, float] = field(init=False)

    def __post_init__(self):
        self.confusion_matrix = defaultdict(dict)
        options = self.obj_cls.possible_instances()
        remainder = (1.0 - self.quality) / (len(options) - 1)
        for truth in options:
            for prediction in options:
                self.confusion_matrix[truth][prediction] = self.quality if truth == prediction else remainder

        self.random_vector = {option: 1 / len(options) for option in options}

    def sample(self, gt: Optional[base.Object] = None, temperature: float = 0.5) -> mapping.Objects:
        values = self.confusion_matrix[gt].values() if gt else self.random_vector.values()
        probs = softmax(np.array(list(values)) / temperature)
        return (self.choice(probs),)

    def choice(self, probs: List[float]) -> base.Object:
        return np.random.choice(list(self.confusion_matrix.keys()), p=probs)


@dataclass
class SbSWorkerConfig(WorkerConfig):
    left_bias: float
    treat_bias_as_original_a: bool = False
    obj_cls: Type[base.SbSChoice] = field(init=False, default=base.SbSChoice)

    def __post_init__(self):
        self.confusion_matrix = defaultdict(dict)
        options = self.obj_cls.possible_instances()
        assert len(options) == 2
        a, b = options
        assert a.value == 'a'
        assert b.value == 'b'

        self.confusion_matrix[a][a] = self.quality + (1.0 - self.quality) * self.left_bias
        self.confusion_matrix[a][b] = (1.0 - self.quality) * (1.0 - self.left_bias)

        self.confusion_matrix[b][a] = (1.0 - self.quality) * self.left_bias
        self.confusion_matrix[b][b] = self.quality + (1.0 - self.quality) * (1.0 - self.left_bias)

        self.random_vector = {a: self.left_bias, b: 1.0 - self.left_bias}


@dataclass
class AnnotationWorkerConfig(WorkerConfig):
    obj_cls: Type[base.Class] = base.BinaryEvaluation

    def sample(
        self,
        gt: Optional[Tuple[mapping.Objects, mapping.Objects, Optional[mapping.Objects]]] = None,
        temperature: float = 0.5,
    ) -> mapping.Objects:
        assert gt, 'No ground truth supplied for annotation - impossible to create answer'
        values = self.confusion_matrix[true].values()
        probs = softmax(np.array(list(values)) / temperature)
        ok = self.choice(probs) == true
        task, solution, _ = gt

        if ok:
            return solution

        return mutate(solution)


@dataclass
class AnnotationCheckWorkerConfig(AnnotationWorkerConfig):
    def sample(
        self,
        gt: Optional[Tuple[mapping.Objects, mapping.Objects, Optional[mapping.Objects]]] = None,
        temperature: float = 0.5,
    ) -> mapping.Objects:
        if gt:
            task, solution, given_solution = gt
            answer = base.BinaryEvaluation(solution == given_solution)

            values = self.confusion_matrix[answer].values()
        else:
            values = self.random_vector.values()
        probs = softmax(np.array(list(values)) / temperature)
        return (self.choice(probs),)


@dataclass(init=False)
class Worker:
    client: NDATolokaClient

    config: WorkerConfig

    worker_info: Dict[str, Any]

    # мы можем хотеть, чтобы он н-р работал и на проверках, и на разметке
    # пока не понятно, как это стоит сделать
    pool_id: str

    task_spec: df_client.PreparedTaskSpec

    solutions: Optional[Dict[mapping.TaskID, TaskInfo]] = field(default_factory=dict)

    time_policy: Callable = None

    _thread: Optional[StoppableThread] = field(init=False)

    def __init__(
        self,
        worker_info: Dict[str, Any],
        config: WorkerConfig,
        pool_id: str,
        task_spec: df_client.PreparedTaskSpec,
        task_duration_hint: timedelta,
        solutions: Optional[Dict[str, Dict[str, Any]]] = None,
    ):
        # todo мб просто post init
        self.client = NDATolokaClient(worker_info['oauth'])
        worker_info['id'] = self.client.resolve_id(worker_info['uid'])['id']
        self.worker_info = worker_info
        self.config = config
        self.pool_id = pool_id
        self.task_spec = task_spec

        self.solutions = solutions
        self.time_policy = lambda task_count: self.config.speed.get_timeout(task_duration_hint * task_count)
        self._thread = None

    @property
    def id(self) -> str:
        return self.worker_info['id']

    @property
    def quality(self) -> float:
        return self.config.quality

    def solve_task(self, task: Dict[str, Any], task_id: mapping.TaskID) -> mapping.Objects:
        if self.solutions is not None:
            info = self.solutions[task_id]
            gt = info.solution
            solution = self.config.sample(gt=gt, temperature=info.complexity)
            logger.debug(f'Created {solution} from gt {gt}')
        else:
            solution = self.config.sample()
            logger.debug(f'Created {solution} from scratch')

        return solution

    def get_target(self) -> Callable:

        # only one loop iteration,
        # loop functionality encased in StoppableThread
        def work() -> bool:

            try:
                assignment = self.client.request_assignment(pool_id=self.pool_id)
                start_time = datetime.now()
                logger.debug(f'Got assignment {assignment} for worker #{id(self)}')
            # пока не очень понятно, что стоит делать в таких случаях
            except IncorrectActionsApiError as e:
                # out of tasks, but maybe they will be uploaded at next loop iteration
                logger.debug(f'Out of tasks for worker #{id(self)}, got {e}')
                return True
            except DoesNotExistApiError as e:
                # out of control tasks for this user, no more work possible
                logger.debug(f'Out of control tasks for worker #{id(self)}, got {e}')
                return True
            except Exception as e:
                logger.debug(f'Got unexpected exception: {e} in worker #{id(self)}')
                return True
            solutions = []
            for task in assignment['tasks']:
                task_id = mapping.TaskID(self.task_spec.task_mapping.from_task_values(task['input_values']))
                logger.debug(f'Got task_id {task_id} for task {task}')

                solutions.append(self.task_spec.task_mapping.toloka_values(self.solve_task(task, task_id), output=True))

            # правая граница не включается
            # мб еще стоит учитывать время, которое потратили на решение задачи,
            # если оно большое
            timeout = self.time_policy(task_count=len(assignment['tasks']))

            # мб можно создавать event для треда извне, и здесь спать на нем,
            # чтобы быстрее выходить - тут не понятно, как лучше
            sleep(timeout)

            # todo - обернуть в retry, если service unavailable, то все развалится и страница останется
            # висеть как минимум до expiration'а
            result = self.client.submit_assignment(assignment['id'], solutions)
            logger.debug(f'Submitted {solutions} after {datetime.now() - start_time}')
            assert result['success'], 'Could not submit assignment'
            return False

        return work

    def start(self):
        assert self._thread is None, 'Not configured'
        self._thread = StoppableThread(target=self.get_target())
        self._thread.start()

    def stop(self):
        self._thread.stop()
        self._thread.join()
        self._thread = None


@dataclass(init=False)
class SbSWorker(Worker):
    def __init__(
        self,
        worker_info: Dict[str, Any],
        config: SbSWorkerConfig,
        pool_id: str,
        task_spec: df_client.PreparedTaskSpec,
        task_duration_hint: timedelta,
        solutions: Optional[Dict[str, Dict[str, Any]]] = None,
    ):
        super(SbSWorker, self).__init__(
            worker_info,
            config,
            pool_id,
            task_spec,
            task_duration_hint,
            solutions,
        )
        self.stats = {'not_swapped': 0, 'swapped': 0}

    def register_task(self, task: Dict[str, Any]) -> bool:
        input_objects = list(
            map(
                content,
                self.task_spec.task_mapping.from_task_values(task['input_values'])[
                    len(self.task_spec.function.hints) :
                ],
            )
        )
        assert input_objects, 'No input objects'
        assert len(input_objects) % 2 == 0, 'Odd number of input objects'
        left_objects, right_objects = input_objects[: len(input_objects) // 2], input_objects[len(input_objects) // 2 :]
        a_left, a_right = 'A' in left_objects[0], 'A' in right_objects[0]
        b_left, b_right = 'B' in left_objects[0], 'B' in right_objects[0]

        assert a_left != a_right
        assert b_left != b_right
        assert a_left == b_right

        for left, right in zip(left_objects, right_objects):
            a_left_loc, a_right_loc = 'A' in left, 'A' in right
            b_left_loc, b_right_loc = 'B' in left, 'B' in right

            assert a_left == a_left_loc
            assert a_right == a_right_loc
            assert b_left == b_left_loc
            assert b_right == b_right_loc

            self.stats['not_swapped'] += int(a_left_loc)
            self.stats['swapped'] += int(a_right_loc)
        return a_right

    def solve_task(self, task: Dict[str, Any], task_id: mapping.TaskID) -> Tuple[base.SbSChoice]:
        swapped = self.register_task(task)
        need_swap = self.config.treat_bias_as_original_a and swapped
        if need_swap:
            swapped_task = swap_task(self.task_spec.task_mapping.from_task_values(task['input_values']), self.task_spec)
            task = self.task_spec.task_mapping.toloka_values(swapped_task)

        answer = super(SbSWorker, self).solve_task(task, task_id)
        if need_swap:
            answer = answer.swap()
        return (answer,)


@dataclass(init=False)
class WorkerGroup:
    workers: List[Worker]

    def __init__(
        self,
        pool_id: str,
        task_spec: df_client.PreparedTaskSpec,
        task_duration_hint: timedelta,
        # WorkerConfig(speed, quality), count
        worker_types: List[Tuple[WorkerConfig, int]],
        solutions: Optional[Dict[str, Dict[str, Any]]] = None,
        worker_cls: Type[Worker] = Worker,
    ):
        thread_pool = ThreadPool(8)
        configs = [config for config, count in worker_types for _ in range(count)]
        clients = [NDARegisterClient() for _ in configs]
        workers_info = thread_pool.map(lambda register_client: register_client.create_user(), clients)
        self.workers = [
            worker_cls(
                worker_info=info,
                pool_id=pool_id,
                task_spec=task_spec,
                task_duration_hint=task_duration_hint,
                solutions=solutions,
                config=config,
            )
            for info, config in zip(workers_info, configs)
        ]
        shuffle(self.workers)

    def start(self):
        for worker in self.workers:
            worker.start()

    def stop(self):
        for worker in self.workers:
            worker.stop()



from dataclasses import dataclass, fields
from enum import Enum
from typing import Dict, Union

DEFAULT_LANG = 'EN'


# ISO 639-1 for language codes, as in Toloka
@dataclass(frozen=True)
class LocalizedString:
    lang_to_text: Dict[str, str]

    def __getitem__(self, lang: str) -> str:
        if lang not in self.lang_to_text:
            lang = DEFAULT_LANG
        assert lang in self.lang_to_text, f'no text for {lang} language'
        return self.lang_to_text[lang]

    def __add__(self, other: Union['LocalizedString', str]) -> 'LocalizedString':
        if isinstance(other, str):
            return LocalizedString({lang: text + other for lang, text in self.lang_to_text.items()})
        return LocalizedString(
            {lang: self[lang] + other[lang] for lang in self.lang_to_text.keys() | other.lang_to_text.keys()}
        )

    @staticmethod
    def convert_data(data: Union[Dict[str, str], 'LocalizedString']) -> 'LocalizedString':
        if isinstance(data, LocalizedString):
            return data
        return LocalizedString(data)

    @staticmethod
    def convert_dataclass(obj: dataclass) -> None:
        for f in fields(obj):
            if f.type == LocalizedString:
                setattr(obj, f.name, LocalizedString.convert_data(getattr(obj, f.name)))


EMPTY_STRING = LocalizedString({DEFAULT_LANG: ''})


class TextFormat(Enum):
    PLAIN = 'plain'
    MARKDOWN = 'markdown'


@dataclass
class Title:
    text: LocalizedString
    format: TextFormat = TextFormat.PLAIN

    def __post_init__(self):
        LocalizedString.convert_dataclass(self)



from dataclasses import dataclass
import logging
from typing import List, Union, Tuple

import toloka.client as toloka

from .. import base, classification_loop, mapping, project as project_config, pool as pool_config, utils


logger = logging.getLogger(__name__)


# scenario automatically changes TaskMapping for PreparedTaskSpec (see builders.py)
# - EXPERT_LABELING_OF_TASKS adds expert meta fields (_ok, _comment) to outputs
# - EXPERT_LABELING_OF_SOLVED_TASKS moves function outputs to inputs, and expert meta fields is outputs
#
# classification function
#     **expert case**     **required fields**                 **arguments**
#     task markup:        inputs | outputs   _ok _comment     [scenario == EXPERT_LABELING_OF_TASKS]
#     solution markup:    inputs   outputs | _ok _comment     [scenario == EXPERT_LABELING_OF_SOLVED_TASKS]
#     verification:       inputs   outputs | _ok _comment     [scenario == EXPERT_LABELING_OF_SOLVED_TASKS]
#
# for annotation function, expected evaluation of annotation is added to expert meta fields, resulting to (eval, _ok, _comment),
# for all expert markup cases except verification case (see AnnotationBuilder in builders.py)
# - for solution markup we use same scenario EXPERT_LABELING_OF_TASKS as for task markup, but we use nested task_spec.check
# - for verification, we do not need eval column, but we need to move outputs to inputs so we use EXPERT_LABELING_OF_SOLVED_TASKS
#
# annotation function
#     **expert case**     **required fields**                     **arguments**
#     task markup:        inputs | outputs   eval _ok _comment    [scenario == EXPERT_LABELING_OF_TASKS]
#     solution markup:    inputs   outputs | eval _ok _comment    [scenario == EXPERT_LABELING_OF_TASKS, task_spec == task_spec.check]
#     verification:       inputs   outputs |      _ok _comment    [scenario == EXPERT_LABELING_OF_SOLVED_TASKS]
#
#     eval column is added to expert meta by AnnotationBuilder
@dataclass
class ExpertPipeline:
    client: toloka.TolokaClient
    task_function: base.TaskFunction
    task_mapping: mapping.TaskMapping
    lang: str
    scenario: project_config.Scenario

    def __post_init__(self):
        task_annotation = self.scenario == project_config.Scenario.EXPERT_LABELING_OF_TASKS
        solution_annotation = self.scenario == project_config.Scenario.EXPERT_LABELING_OF_SOLVED_TASKS
        annotation_verification = False
        if isinstance(self.task_function, base.AnnotationFunction):
            # for task labeling, we have task function outputs,
            # expert annotation fields and extra evaluation field in output mapping
            task_annotation = (
                len(self.task_mapping.output_mapping)
                == len(self.task_function.outputs) + len(project_config.Builder.EXPERT_LABELING_METAS) + 1
            )

            # for solution annotation, we only have expert annotation fields and evaluation in output mapping
            solution_annotation = (
                len(self.task_mapping.output_mapping) == len(project_config.Builder.EXPERT_LABELING_METAS) + 1
            )

            annotation_verification = len(self.task_mapping.output_mapping) == len(
                project_config.Builder.EXPERT_LABELING_METAS
            )

        assert annotation_verification or task_annotation or solution_annotation, 'Incorrect configuration'
        self.scenario = (
            project_config.Scenario.EXPERT_LABELING_OF_TASKS
            if task_annotation
            else project_config.Scenario.EXPERT_LABELING_OF_SOLVED_TASKS
        )

    def create_pool(
        self,
        pool_cfg: pool_config.ExpertConfig,
    ) -> toloka.Pool:
        pool = self.client.create_pool(pool_config.create_expert_pool_params(pool_cfg))
        return pool

    def add_input_objects(
        self, pool_id: str, input_objects: Union[List[mapping.TaskSingleSolution], List[mapping.Objects]]
    ):
        assert input_objects, 'No objects supplied'
        # todo not very reliable typecheck due to dynamic generics
        # todo: verification case for annotations

        solution_annotation = (
            isinstance(input_objects[0], tuple)
            and len(input_objects[0]) == 2
            and isinstance(input_objects[0][0], tuple)
            and isinstance(input_objects[0][1], tuple)
        )
        tasks = []
        for task_objects_or_solutions in input_objects:
            if solution_annotation:
                task_objects_or_solutions = task_objects_or_solutions[0] + task_objects_or_solutions[1]
            self.task_mapping.validate_objects(task_objects_or_solutions)
            task = self.task_mapping.to_task(task_objects_or_solutions)
            task.pool_id = pool_id
            tasks.append(task)
        logger.debug(f'creating {len(tasks)} tasks')
        self.client.create_tasks(tasks, allow_defaults=True, async_mode=True, skip_invalid_items=False)
        self.client.open_pool(pool_id)

    def loop(self, pool_id: str):
        utils.wait_pool_for_close(self.client, pool_id)

    def get_results(
        self,
        pool_id: str,
        objects: Union[List[mapping.TaskSingleSolution], List[mapping.Objects]],
    ) -> List[Tuple[Union[mapping.TaskSingleSolution, mapping.Objects], mapping.TaskMultipleSolutions]]:
        assignments = [
            assignment
            for assignment, _ in classification_loop.get_assignments_solutions(
                self.client, self.task_mapping, pool_id, [toloka.Assignment.ACCEPTED]
            )
        ]
        if self.scenario == project_config.Scenario.EXPERT_LABELING_OF_SOLVED_TASKS:
            objects = [obj[0] + obj[1] for obj in objects]
        return list(zip(objects, mapping.get_solutions(assignments, self.task_mapping, objects)))



import abc
import datetime
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
import json
from typing import Any, Callable, Dict, Generic, List, Optional, Type, TypeVar, Union

from pure_protobuf.dataclasses_ import field, message, one_of, part
from pure_protobuf.oneof import OneOf_
import toloka.client as toloka

from .. import base, classification, classification_loop, client, control, evaluation, objects, pricing, worker
from ..utils import DecimalEncoder

# We use Protobuf to serialize objects on lzy whiteboards.
# Standard pickle serialization does not fit our needs, because in general case refactorings in library code may lead
# to incorrect objects deserialization for old, already persisted whiteboards.
#
# For each class which will be stored in whiteboards, we define its serialization wrapper. A possible alternative is
# to provide additional metadata (@message class decorator, attributes data, etc.) to source classes, but we fully
# separate serialization logic from pure business-logic classes, because:
# 1) Currently we have issues with `pure-protobuf` dependency in Arcadia contrib.
# 2) Some attributes types, i.e. dicts, are not Protobuf serializable, and we don't want to modify source classes only
#    because of serialization needs.
#
# To serialize classes from inheritance schemes, we use Protobuf one_of.
#
# We also have another case, when client defines his own classes. Such sets of classes are limited (i.e. subclasses of
# Class) and we typically know data of these classes (.value for Class inheritor). But we don't know to which class
# we need to deserialize this data. So we introduce classes registry, and the client have to associate a permanent
# name with each of his self-defined class, so we can later instantiate this concrete class by its name.

TolokaObj = toloka.primitives.base.BaseTolokaObject

T = TypeVar('T', bound=Type[Any], covariant=True)
ClassT = TypeVar('ClassT', bound=Type[base.Class], covariant=True)
TolokaObjT = TypeVar('TolokaObjT', bound=Type[TolokaObj], covariant=True)

# DO NOT CHANGE THIS NAMES!
# They are persisted in lzy whiteboards and must remain the same for correct deserialization.
types_registry: Dict[Type[base.Object], str] = {
    base.SbSChoice: 'SbSChoice',
    base.BinaryEvaluation: 'BinaryEvaluation',
    objects.Audio: 'Audio',
    objects.Image: 'Image',
    objects.Text: 'Text',
    objects.Video: 'Video',
}
types_registry_reversed = {name: type for type, name in types_registry.items()}


def register_type(type: Type[base.Object], name: str):
    global types_registry, types_registry_reversed
    types_registry[type] = name
    types_registry_reversed[name] = type


func_registry = {}
func_registry_reversed = {}


def register_func(func: Callable, name: str):
    global func_registry, func_registry_reversed
    func_registry[func] = name
    func_registry_reversed[name] = func


# Most likely we won't need exact version of function as it was implemented at the time of the launch.
# Its outputs will be persisted in whiteboard anyway. So if we can't find it, it's not a problem.
def func_not_found(name: str) -> Callable:
    def f():
        raise RuntimeError(f'function "{name}" not found')

    return f


def load_func(name: str) -> Callable:
    return func_registry_reversed.get(name, func_not_found(name))


def deserialize_one_of_field(field):
    one_of = getattr(field, field.which_one_of)
    if isinstance(one_of, ProtobufSerializer):
        return one_of.deserialize()
    else:
        return one_of  # primitive type


class ProtobufSerializer(Generic[T]):
    @staticmethod
    @abc.abstractmethod
    def serialize(obj: T) -> 'ProtobufSerializer':
        ...

    @abc.abstractmethod
    def deserialize(self) -> T:
        ...


@message
@dataclass
class LocalizedStringEntry:
    lang: str = field(1)
    text: str = field(2)


@message
@dataclass
class LocalizedString(ProtobufSerializer[base.LocalizedString]):
    entries: List[LocalizedStringEntry] = field(1)

    @staticmethod
    def serialize(obj: base.LocalizedString) -> 'LocalizedString':
        return LocalizedString(entries=[LocalizedStringEntry(lang, text) for lang, text in obj.lang_to_text.items()])

    def deserialize(self) -> base.LocalizedString:
        return base.LocalizedString(lang_to_text={entry.lang: entry.text for entry in self.entries})


# For library string enumerations
@message
@dataclass
class StrEnum(ProtobufSerializer[Enum], Generic[ClassT]):
    value: str = field(1)

    @staticmethod
    @abc.abstractmethod
    def enum_cls() -> Type[Enum]:
        ...

    @classmethod
    def serialize(cls, obj: Enum) -> 'StrEnum':
        return cls(value=obj.value)

    def deserialize(self) -> Enum:
        return self.enum_cls()(self.value)


@message
class TextFormat(StrEnum[base.TextFormat]):
    @staticmethod
    def enum_cls() -> Type[base.TextFormat]:
        return base.TextFormat


@message
@dataclass
class Title(ProtobufSerializer[base.Title]):
    text: LocalizedString = field(1)
    format: TextFormat = field(2)

    @staticmethod
    def serialize(obj: base.Title) -> 'Title':
        return Title(text=LocalizedString.serialize(obj.text), format=TextFormat.serialize(obj.format))

    def deserialize(self) -> base.Title:
        return base.Title(text=self.text.deserialize(), format=self.format.deserialize())


@message
class SbSChoice(StrEnum[base.SbSChoice]):
    @staticmethod
    def enum_cls() -> Type[base.SbSChoice]:
        return base.SbSChoice


@message
@dataclass
class BinaryEvaluation(ProtobufSerializer[base.BinaryEvaluation]):
    ok: bool = field(1)

    @staticmethod
    def serialize(obj: base.BinaryEvaluation) -> 'BinaryEvaluation':
        return BinaryEvaluation(obj.ok)

    def deserialize(self) -> base.BinaryEvaluation:
        return base.BinaryEvaluation(self.ok)


@message
@dataclass
class Metadata(ProtobufSerializer[base.Metadata]):
    metadata: str = field(1)

    @staticmethod
    def serialize(obj: base.Metadata) -> 'Metadata':
        return Metadata(obj.metadata)

    def deserialize(self) -> base.Metadata:
        return base.Metadata(self.metadata)


@message
@dataclass
class Text(ProtobufSerializer[objects.Text]):
    text: str = field(1)

    @staticmethod
    def serialize(obj: objects.Text) -> 'Text':
        return Text(obj.text)

    def deserialize(self) -> objects.Text:
        return objects.Text(self.text)


@message
@dataclass
class Audio(ProtobufSerializer[objects.Audio]):
    url: str = field(1)

    @staticmethod
    def serialize(obj: objects.Audio) -> 'Audio':
        return Audio(obj.url)

    def deserialize(self) -> objects.Audio:
        return objects.Audio(self.url)


@message
@dataclass
class Image(ProtobufSerializer[objects.Image]):
    url: str = field(1)

    @staticmethod
    def serialize(obj: objects.Image) -> 'Image':
        return Image(obj.url)

    def deserialize(self) -> objects.Image:
        return objects.Image(self.url)


@message
@dataclass
class Video(ProtobufSerializer[objects.Video]):
    url: str = field(1)

    @staticmethod
    def serialize(obj: objects.Video) -> 'Video':
        return Video(obj.url)

    def deserialize(self) -> objects.Video:
        return objects.Video(self.url)


# For user-defined string enumerations, including Class inheritors
@message
@dataclass
class Class(ProtobufSerializer[base.Class]):
    value: str = field(1)
    cls: str = field(2)

    @staticmethod
    def serialize(obj: base.Class) -> 'Class':
        cls = types_registry[type(obj)]
        return Class(value=obj.value, cls=cls)

    def deserialize(self) -> base.Class:
        cls = types_registry_reversed[self.cls]
        return cls(self.value)


@message
@dataclass
class Label(ProtobufSerializer[base.Label]):
    label: OneOf_ = one_of(
        cls=part(Class, 1),
        sbs_choice=part(SbSChoice, 2),
        binary_evaluation=part(BinaryEvaluation, 3),
    )

    @staticmethod
    def serialize(obj: base.Label) -> 'Label':
        label = Label()
        if isinstance(obj, base.BinaryEvaluation):
            label.label.binary_evaluation = BinaryEvaluation.serialize(obj)
        elif isinstance(obj, base.SbSChoice):
            label.label.sbs_choice = SbSChoice.serialize(obj)
        elif isinstance(obj, base.Class):
            label.label.cls = Class.serialize(obj)
        else:
            assert False, f'unexpected Label type: {type(obj)}'
        return label

    def deserialize(self) -> base.Label:
        return deserialize_one_of_field(self.label)


@message
@dataclass
class Object(ProtobufSerializer[base.Object]):
    obj: OneOf_ = one_of(
        cls=part(Class, 1),
        sbs_choice=part(SbSChoice, 2),
        binary_evaluation=part(BinaryEvaluation, 3),
        metadata=part(Metadata, 4),
        text=part(Text, 5),
        audio=part(Audio, 6),
        image=part(Image, 7),
        video=part(Video, 8),
    )

    @staticmethod
    def serialize(obj: base.Object) -> 'Object':
        obj_ = Object()
        if isinstance(obj, base.BinaryEvaluation):
            obj_.obj.binary_evaluation = BinaryEvaluation.serialize(obj)
        elif isinstance(obj, base.SbSChoice):
            obj_.obj.sbs_choice = SbSChoice.serialize(obj)
        elif isinstance(obj, base.Class):
            obj_.obj.cls = Class.serialize(obj)
        elif isinstance(obj, base.Metadata):
            obj_.obj.metadata = Metadata.serialize(obj)
        elif isinstance(obj, objects.Text):
            obj_.obj.text = Text.serialize(obj)
        elif isinstance(obj, objects.Audio):
            obj_.obj.audio = Audio.serialize(obj)
        elif isinstance(obj, objects.Image):
            obj_.obj.image = Image.serialize(obj)
        elif isinstance(obj, objects.Video):
            obj_.obj.video = Video.serialize(obj)
        else:
            assert False, f'unexpected Object type: {type(obj)}'
        return obj_

    def deserialize(self) -> base.Object:
        return deserialize_one_of_field(self.obj)


@message
@dataclass
class ObjectMeta(ProtobufSerializer[base.ObjectMeta]):
    type: str = field(1)
    name: Optional[str] = field(2, default=None)
    title: Optional[Title] = field(3, default=None)
    required: bool = field(4, default=True)

    @staticmethod
    def serialize(obj: base.ObjectMeta) -> 'ObjectMeta':
        return ObjectMeta(
            type=types_registry[obj.type],
            name=obj.name,
            title=Title.serialize(obj.title) if obj.title else None,
            required=obj.required,
        )

    def deserialize(self) -> base.ObjectMeta:
        return base.ObjectMeta(
            type=types_registry_reversed[self.type],
            name=self.name,
            title=self.title.deserialize() if self.title else None,
            required=self.required,
        )


@message
@dataclass
class AvailableLabels(ProtobufSerializer[base.AvailableLabels]):
    labels: List[Label] = field(1)

    @staticmethod
    def serialize(obj: base.AvailableLabels) -> 'AvailableLabels':
        return AvailableLabels([Label.serialize(label) for label in obj.labels])

    def deserialize(self) -> base.AvailableLabels:
        return base.AvailableLabels([label.deserialize() for label in self.labels])


@message
@dataclass
class Consequence(ProtobufSerializer[base.Consequence]):
    consequence: OneOf_ = one_of(
        available_labels=part(AvailableLabels, 1),
    )

    @staticmethod
    def serialize(obj: base.Consequence) -> 'Consequence':
        consequence = Consequence()
        if isinstance(obj, base.AvailableLabels):
            consequence.consequence.available_labels = AvailableLabels.serialize(obj)
        else:
            raise ValueError(f'unexpected consequence type: {type(obj)}')
        return consequence

    def deserialize(self) -> base.Consequence:
        return deserialize_one_of_field(self.consequence)


@message
@dataclass
class ConditionEquals(ProtobufSerializer[base.ConditionEquals]):
    what: str = field(1)
    to: OneOf_ = one_of(
        str=part(str, 2),
    )

    @staticmethod
    def serialize(obj: base.ConditionEquals) -> 'ConditionEquals':
        condition = ConditionEquals(what=obj.what)
        if isinstance(obj.to, str):
            condition.to.str = obj.to
        else:
            raise ValueError(f'unsupported equals to type: {type(obj.to)}')
        return condition

    def deserialize(self) -> base.ConditionEquals:
        return base.ConditionEquals(what=self.what, to=deserialize_one_of_field(self.to))


@message
@dataclass
class Condition(ProtobufSerializer[base.Condition]):
    condition: OneOf_ = one_of(
        equals=part(ConditionEquals, 1),
    )

    @staticmethod
    def serialize(obj: base.Condition) -> 'Condition':
        condition = Condition()
        if isinstance(obj, base.ConditionEquals):
            condition.condition.equals = ConditionEquals.serialize(obj)
        else:
            raise ValueError(f'unexpected condition type: {type(obj)}')
        return condition

    def deserialize(self) -> base.Condition:
        return deserialize_one_of_field(self.condition)


# then/else branches are originally unions, better option to pack them to one_of, but it's not possible because union
# contains If itself and we can't pass class to one_of part using quotation syntax like 'If'
@dataclass
class If(ProtobufSerializer[base.If]):
    condition: Condition = field(1)
    then_if: Optional['If'] = field(2, default=None)
    then_consequence: Optional[Consequence] = field(3, default=None)
    else_if: Optional['If'] = field(4, default=None)
    else_consequence: Optional[Consequence] = field(5, default=None)

    def __post_init__(self):
        assert (self.then_if is not None) ^ (self.then_consequence is not None)
        assert not (self.else_if is not None and self.else_consequence is not None)

    @staticmethod
    def serialize(obj: base.If) -> 'If':
        return If(
            condition=Condition.serialize(obj.condition),
            then_if=If.serialize(obj.then) if obj.then and isinstance(obj.then, base.If) else None,
            then_consequence=Consequence.serialize(obj.then)
            if obj.then and isinstance(obj.then, base.Consequence)
            else None,
            else_if=If.serialize(obj.else_) if obj.else_ and isinstance(obj.else_, base.If) else None,
            else_consequence=Consequence.serialize(obj.else_)
            if obj.else_ and isinstance(obj.else_, base.Consequence)
            else None,
        )

    def deserialize(self) -> base.If:
        then = self.then_if.deserialize() if self.then_if else self.then_consequence.deserialize()
        else_ = None
        if self.else_if:
            else_ = self.else_if.deserialize()
        if self.else_consequence:
            else_ = self.else_consequence.deserialize()
        return base.If(condition=self.condition.deserialize(), then=then, else_=else_)


# self-referencing classes are decorated with @message separately, see https://github.com/eigenein/protobuf/issues/96
If = message(If)


@message
class LabelsDisplayType(StrEnum[base.LabelsDisplayType]):
    @staticmethod
    def enum_cls() -> Type[base.LabelsDisplayType]:
        return base.LabelsDisplayType


@message
@dataclass
class ClassMeta(ObjectMeta, ProtobufSerializer[base.ClassMeta]):
    available_labels: Optional[If] = field(100, default=None)
    input_display_type: LabelsDisplayType = field(101, default=LabelsDisplayType(base.LabelsDisplayType.MULTI))
    text_format: TextFormat = field(102, default=TextFormat(base.TextFormat.PLAIN))

    @staticmethod
    def serialize(obj: base.ClassMeta) -> 'ClassMeta':
        obj_meta = ObjectMeta.serialize(obj)
        return ClassMeta(
            type=obj_meta.type,
            name=obj_meta.name,
            title=obj_meta.title,
            required=obj_meta.required,
            available_labels=If.serialize(obj.available_labels) if obj.available_labels else None,
            input_display_type=LabelsDisplayType.serialize(obj.input_display_type),
            text_format=TextFormat.serialize(obj.text_format),
        )

    def deserialize(self) -> base.ClassMeta:
        obj_meta = super(ClassMeta, self).deserialize()
        return base.ClassMeta(
            type=obj_meta.type,
            name=obj_meta.name,
            title=obj_meta.title,
            required=obj_meta.required,
            available_labels=self.available_labels.deserialize() if self.available_labels else None,
            input_display_type=self.input_display_type.deserialize(),
            text_format=self.text_format.deserialize(),
        )


@message
@dataclass
class TextValidation(ProtobufSerializer[objects.TextValidation]):
    regex: LocalizedString = field(1)
    hint: LocalizedString = field(2)

    @staticmethod
    def serialize(obj: objects.TextValidation) -> 'TextValidation':
        return TextValidation(
            regex=LocalizedString.serialize(obj.regex),
            hint=LocalizedString.serialize(obj.hint),
        )

    def deserialize(self) -> objects.TextValidation:
        return objects.TextValidation(
            regex=self.regex.deserialize(),
            hint=self.hint.deserialize(),
        )


@message
@dataclass
class TextMeta(ObjectMeta, ProtobufSerializer[objects.TextMeta]):
    format: TextFormat = field(100, default=TextFormat(base.TextFormat.PLAIN))
    validation: Optional[TextValidation] = field(101, default=None)

    @staticmethod
    def serialize(obj: objects.TextMeta) -> 'TextMeta':
        obj_meta = ObjectMeta.serialize(obj)
        return TextMeta(
            type=obj_meta.type,
            name=obj_meta.name,
            title=obj_meta.title,
            required=obj_meta.required,
            format=TextFormat.serialize(obj.format),
            validation=TextValidation.serialize(obj.validation) if obj.validation else None,
        )

    def deserialize(self) -> objects.TextMeta:
        obj_meta = super(TextMeta, self).deserialize()
        return objects.TextMeta(
            type=obj_meta.type,
            name=obj_meta.name,
            title=obj_meta.title,
            required=obj_meta.required,
            format=self.format.deserialize(),
            validation=self.validation.deserialize() if self.validation else None,
        )


@message
@dataclass
class ObjectMetaT(ProtobufSerializer[base.ObjectMeta]):
    meta: OneOf_ = one_of(
        object=part(ObjectMeta, 1),
        cls=part(ClassMeta, 2),
        text=part(TextMeta, 3),
    )

    @staticmethod
    def serialize(obj: base.ObjectMeta) -> 'ObjectMetaT':
        meta = ObjectMetaT()
        if isinstance(obj, base.ClassMeta):
            meta.meta.cls = ClassMeta.serialize(obj)
        elif isinstance(obj, objects.TextMeta):
            meta.meta.text = TextMeta.serialize(obj)
        else:
            meta.meta.object = ObjectMeta.serialize(obj)
        return meta

    def deserialize(self) -> base.ObjectMeta:
        return deserialize_one_of_field(self.meta)


@message
@dataclass
class FunctionArgument(ProtobufSerializer[Union[Type[base.Object], base.ObjectMeta]]):
    arg: OneOf_ = one_of(
        type=part(str, 1),
        meta=part(ObjectMetaT, 2),
    )

    @staticmethod
    def serialize(obj: Union[Type[base.Object], base.ObjectMeta]) -> 'FunctionArgument':
        arg = FunctionArgument()
        if isinstance(obj, base.ObjectMeta):
            arg.arg.meta = ObjectMetaT.serialize(obj)
        else:
            arg.arg.type = types_registry[obj]
        return arg

    def deserialize(self) -> Union[Type[base.Object], base.ObjectMeta]:
        which_one = self.arg.which_one_of
        if which_one == 'type':
            return types_registry_reversed[self.arg.type]
        else:
            return self.arg.meta.deserialize()


@message
@dataclass
class ClassificationFunction(ProtobufSerializer[base.ClassificationFunction]):
    inputs: List[FunctionArgument] = field(1)
    cls: FunctionArgument = field(2)

    @staticmethod
    def serialize(obj: base.ClassificationFunction) -> 'ClassificationFunction':
        return ClassificationFunction(
            inputs=[FunctionArgument.serialize(input) for input in obj.inputs],
            cls=FunctionArgument.serialize(obj.cls),
        )

    def deserialize(self) -> base.ClassificationFunction:
        return base.ClassificationFunction(
            inputs=tuple(FunctionArgument.deserialize(input) for input in self.inputs),
            cls=FunctionArgument.deserialize(self.cls),
        )


sbs_choice = FunctionArgument()
sbs_choice.arg.type = types_registry[base.SbSChoice]


@message
@dataclass
class SbSFunction(ProtobufSerializer[base.SbSFunction]):
    inputs: List[FunctionArgument] = field(1)
    hints: List[FunctionArgument] = field(2, default_factory=list)
    choice: FunctionArgument = field(3, default=sbs_choice)

    @staticmethod
    def serialize(obj: base.SbSFunction) -> 'SbSFunction':
        return SbSFunction(
            inputs=[FunctionArgument.serialize(input) for input in obj.inputs],
            hints=[FunctionArgument.serialize(hint) for hint in obj.hints] if obj.hints else [],
            choice=FunctionArgument.serialize(obj.choice),
        )

    def deserialize(self) -> base.SbSFunction:
        return base.SbSFunction(
            inputs=tuple(FunctionArgument.deserialize(input) for input in self.inputs),
            hints=tuple(FunctionArgument.deserialize(hint) for hint in self.hints) if self.hints else None,
            choice=FunctionArgument.deserialize(self.choice),
        )


@message
@dataclass
class AnnotationFunction(ProtobufSerializer[base.AnnotationFunction]):
    inputs: List[FunctionArgument] = field(1)
    outputs: List[FunctionArgument] = field(2)
    evaluation: FunctionArgument = field(3)

    @staticmethod
    def serialize(obj: base.AnnotationFunction) -> 'AnnotationFunction':
        return AnnotationFunction(
            inputs=[FunctionArgument.serialize(input) for input in obj.inputs],
            outputs=[FunctionArgument.serialize(output) for output in obj.outputs],
            evaluation=FunctionArgument.serialize(obj.evaluation),
        )

    def deserialize(self) -> base.AnnotationFunction:
        return base.AnnotationFunction(
            inputs=tuple(FunctionArgument.deserialize(input) for input in self.inputs),
            outputs=tuple(FunctionArgument.deserialize(output) for output in self.outputs),
            evaluation=FunctionArgument.deserialize(self.evaluation),
        )


@message
@dataclass
class TaskFunction(ProtobufSerializer[base.TaskFunction]):
    function: OneOf_ = one_of(
        classification=part(ClassificationFunction, 1),
        sbs=part(SbSFunction, 2),
        annotation=part(AnnotationFunction, 3),
    )

    @staticmethod
    def serialize(obj: base.TaskFunction) -> 'TaskFunction':
        function = TaskFunction()
        if isinstance(obj, base.ClassificationFunction):
            function.function.classification = ClassificationFunction.serialize(obj)
        elif isinstance(obj, base.SbSFunction):
            function.function.sbs = SbSFunction.serialize(obj)
        elif isinstance(obj, base.AnnotationFunction):
            function.function.annotation = AnnotationFunction.serialize(obj)
        else:
            raise ValueError(f'unexpected function {obj}')
        return function

    def deserialize(self) -> base.TaskFunction:
        return deserialize_one_of_field(self.function)


@message
@dataclass
class TaskSpec(ProtobufSerializer[base.TaskSpec]):
    id: str = field(1)
    function: TaskFunction = field(2)
    name: LocalizedString = field(3)
    description: LocalizedString = field(4)
    instruction: LocalizedString = field(5)

    @staticmethod
    def serialize(obj: base.TaskSpec) -> 'TaskSpec':
        return TaskSpec(
            id=obj.id,
            function=TaskFunction.serialize(obj.function),
            name=LocalizedString.serialize(obj.name),
            description=LocalizedString.serialize(obj.description),
            instruction=LocalizedString.serialize(obj.instruction),
        )

    def deserialize(self) -> base.TaskSpec:
        return base.TaskSpec(
            id=self.id,
            function=self.function.deserialize(),
            name=self.name.deserialize(),
            description=self.description.deserialize(),
            instruction=self.instruction.deserialize(),
        )


@message
class ComparisonType(StrEnum[control.ComparisonType]):
    @staticmethod
    def enum_cls() -> Type[control.ComparisonType]:
        return control.ComparisonType


@message
@dataclass
class PredicateValue(ProtobufSerializer[Union[float, datetime.timedelta]]):
    value: OneOf_ = one_of(
        float_=part(float, 1),
        timedelta=part(datetime.timedelta, 2),
    )

    @staticmethod
    def serialize(obj: Union[float, datetime.timedelta]) -> 'PredicateValue':
        value = PredicateValue()
        if isinstance(obj, datetime.timedelta):
            value.value.timedelta = obj
        else:
            value.value.float_ = obj
        return value

    def deserialize(self) -> Union[float, datetime.timedelta]:
        return deserialize_one_of_field(self.value)


@message
@dataclass
class ThresholdComparisonPredicate(ProtobufSerializer[control.ThresholdComparisonPredicate]):
    threshold: PredicateValue = field(1)
    comparison: ComparisonType = field(2)

    @staticmethod
    def serialize(obj: control.ThresholdComparisonPredicate) -> 'ThresholdComparisonPredicate':
        return ThresholdComparisonPredicate(
            threshold=PredicateValue.serialize(obj.threshold),
            comparison=ComparisonType.serialize(obj.comparison),
        )

    def deserialize(self) -> control.ThresholdComparisonPredicate:
        return control.ThresholdComparisonPredicate(
            threshold=self.threshold.deserialize(),
            comparison=self.comparison.deserialize(),
        )


@message
class BooleanOperator(StrEnum[control.BooleanOperator]):
    @staticmethod
    def enum_cls() -> Type[control.BooleanOperator]:
        return control.BooleanOperator


# better option is to pack predicate types to one_of, but it's not possible because expression predicate
# contains list of Predicates itself, but we can't pass class to one_of part using quotation syntax like 'Predicate'
@dataclass
class Predicate(ProtobufSerializer[control.Predicate]):
    assignment_accuracy: Optional[ThresholdComparisonPredicate] = field(1, default=None)
    assignment_duration: Optional[ThresholdComparisonPredicate] = field(2, default=None)
    always_true: bool = field(3, default=False)
    expression_boolean_operator: Optional[BooleanOperator] = field(4, default=None)
    expression_predicates: List['Predicate'] = field(5, default_factory=list)

    @staticmethod
    def serialize(obj: control.Predicate) -> 'Predicate':
        predicate = Predicate()
        if isinstance(obj, control.AssignmentAccuracyPredicate):
            predicate.assignment_accuracy = ThresholdComparisonPredicate.serialize(obj)
        elif isinstance(obj, control.AssignmentDurationPredicate):
            predicate.assignment_duration = ThresholdComparisonPredicate.serialize(obj)
        elif isinstance(obj, control.AlwaysTruePredicate):
            predicate.always_true = True
        elif isinstance(obj, control.PredicateExpression):
            predicate.expression_boolean_operator = BooleanOperator.serialize(obj.boolean_operator)
            predicate.expression_predicates = [Predicate.serialize(p) for p in obj.predicates]
        else:
            raise ValueError(f'unexpected predicate type: {type(obj)}')
        return predicate

    def deserialize(self) -> control.Predicate:
        threshold_proto = self.assignment_accuracy or self.assignment_duration
        if threshold_proto:
            threshold = ThresholdComparisonPredicate.deserialize(threshold_proto)
            type = (
                control.AssignmentAccuracyPredicate if self.assignment_accuracy else control.AssignmentDurationPredicate
            )
            return type(threshold=threshold.threshold, comparison=threshold.comparison)

        if self.always_true:
            return control.AlwaysTruePredicate()

        assert self.expression_boolean_operator and self.expression_predicates

        return control.PredicateExpression(
            boolean_operator=BooleanOperator.deserialize(self.expression_boolean_operator),
            predicates=[Predicate.deserialize(p) for p in self.expression_predicates],
        )


Predicate = message(Predicate)


@message
@dataclass
class BlockUser(ProtobufSerializer[control.BlockUser]):
    scope: str = field(1)
    private_comment: str = field(2)
    duration: Optional[datetime.timedelta] = field(3, default=None)

    @staticmethod
    def serialize(obj: control.BlockUser) -> 'BlockUser':
        return BlockUser(
            scope=obj.scope.value,
            private_comment=obj.private_comment,
            duration=obj.duration,
        )

    def deserialize(self) -> control.BlockUser:
        return control.BlockUser(
            scope=toloka.user_restriction.UserRestriction.Scope(self.scope),
            private_comment=self.private_comment,
            duration=self.duration,
        )


@message
@dataclass
class GiveBonusToUser(ProtobufSerializer[control.GiveBonusToUser]):
    amount_usd: float = field(1)

    @staticmethod
    def serialize(obj: control.GiveBonusToUser) -> 'GiveBonusToUser':
        return GiveBonusToUser(amount_usd=obj.amount_usd)

    def deserialize(self) -> control.GiveBonusToUser:
        return control.GiveBonusToUser(amount_usd=self.amount_usd)


@message
@dataclass
class SetAssignmentStatus(ProtobufSerializer[control.SetAssignmentStatus]):
    status: str = field(1)

    @staticmethod
    def serialize(obj: control.SetAssignmentStatus) -> 'SetAssignmentStatus':
        return SetAssignmentStatus(status=obj.status.value)

    def deserialize(self) -> control.SetAssignmentStatus:
        return control.SetAssignmentStatus(status=toloka.Assignment.Status(self.status))


@message
@dataclass
class Action(ProtobufSerializer[control.Action]):
    action: OneOf_ = one_of(
        block_user=part(BlockUser, 1),
        give_bonus_to_user=part(GiveBonusToUser, 2),
        set_assignment_status=part(SetAssignmentStatus, 3),
    )

    @staticmethod
    def serialize(obj: control.Action) -> 'Action':
        action = Action()
        if isinstance(obj, control.BlockUser):
            action.action.block_user = BlockUser.serialize(obj)
        elif isinstance(obj, control.GiveBonusToUser):
            action.action.give_bonus_to_user = GiveBonusToUser.serialize(obj)
        elif isinstance(obj, control.SetAssignmentStatus):
            action.action.set_assignment_status = SetAssignmentStatus.serialize(obj)
        else:
            raise ValueError(f'unexpected action type: {type(obj)}')
        return action

    def deserialize(self) -> control.Action:
        return deserialize_one_of_field(self.action)


@message
@dataclass
class Rule(ProtobufSerializer[control.Rule]):
    predicate: Predicate = field(1)
    action: Action = field(2)

    @staticmethod
    def serialize(obj: control.Rule) -> 'Rule':
        return Rule(predicate=Predicate.serialize(obj.predicate), action=Action.serialize(obj.action))

    def deserialize(self) -> control.Rule:
        return control.Rule(predicate=self.predicate.deserialize(), action=self.action.deserialize())


@message
@dataclass
class Control(ProtobufSerializer[control.Control]):
    rules: List[Rule] = field(1, default_factory=list)

    @staticmethod
    def serialize(obj: control.Control) -> 'Control':
        return Control([Rule.serialize(rule) for rule in obj.rules])

    def deserialize(self) -> control.Control:
        return control.Control([rule.deserialize() for rule in self.rules])


@message
@dataclass
class TolokaObject(ProtobufSerializer[TolokaObj], Generic[TolokaObjT]):
    json: str = field(1)

    @classmethod
    def serialize(cls, obj: TolokaObj):
        return cls(json=json.dumps(obj.unstructure(), cls=DecimalEncoder))

    def deserialize(self) -> TolokaObj:
        return self.toloka_cls().structure(json.loads(self.json, parse_float=Decimal))

    @staticmethod
    @abc.abstractmethod
    def toloka_cls() -> Type[TolokaObj]:
        ...


@message
class TolokaPool(TolokaObject[toloka.Pool]):
    @staticmethod
    def toloka_cls() -> Type[TolokaObj]:
        return toloka.Pool


@message
class TolokaFilterCondition(TolokaObject[toloka.filter.FilterCondition]):
    @staticmethod
    def toloka_cls() -> Type[TolokaObj]:
        return toloka.filter.FilterCondition


@message
class TolokaSkill(TolokaObject[toloka.Skill]):
    @staticmethod
    def toloka_cls() -> Type[TolokaObj]:
        return toloka.Skill


@message
@dataclass
class ExpertFilter(ProtobufSerializer[worker.ExpertFilter]):
    skills: List[TolokaSkill] = field(1)

    @staticmethod
    def serialize(obj: worker.ExpertFilter) -> 'ExpertFilter':
        return ExpertFilter([TolokaSkill.serialize(skill) for skill in obj.skills])

    def deserialize(self) -> worker.ExpertFilter:
        return worker.ExpertFilter([skill.deserialize() for skill in self.skills])


@message
@dataclass
class LanguageRequirement(ProtobufSerializer[worker.LanguageRequirement]):
    lang: str = field(1)
    verified: bool = field(2)

    @staticmethod
    def serialize(obj: worker.LanguageRequirement) -> 'LanguageRequirement':
        return LanguageRequirement(lang=obj.lang, verified=obj.verified)

    def deserialize(self) -> worker.LanguageRequirement:
        return worker.LanguageRequirement(lang=self.lang, verified=self.verified)


@message
@dataclass
class BaseWorkerFilter(ProtobufSerializer[worker.BaseWorkerFilter]):
    training_score: Optional[int] = field(1, default=None)

    @staticmethod
    def serialize(obj: worker.BaseWorkerFilter) -> 'BaseWorkerFilter':
        return BaseWorkerFilter(training_score=obj.training_score)

    def deserialize(self) -> worker.BaseWorkerFilter:
        return worker.BaseWorkerFilter(training_score=self.training_score)


@message
@dataclass
class WorkerFilterParams(ProtobufSerializer[worker.WorkerFilter.Params]):
    lang: Optional[LanguageRequirement] = field(1, default=None)
    regions: List[int] = field(2, default_factory=list)
    age_min: Optional[int] = field(3, default=None)
    age_max: Optional[int] = field(4, default=None)
    client_types: List[str] = field(5, default_factory=list)

    @staticmethod
    def serialize(obj: worker.WorkerFilter.Params) -> 'WorkerFilterParams':
        return WorkerFilterParams(
            lang=LanguageRequirement.serialize(obj.lang) if obj.lang else None,
            regions=list(obj.regions),
            age_min=obj.age_range[0],
            age_max=obj.age_range[1],
            client_types=[t.value for t in obj.client_types],
        )

    def deserialize(self) -> worker.WorkerFilter.Params:
        return worker.WorkerFilter.Params(
            lang=self.lang.deserialize() if self.lang else None,
            regions=set(self.regions),
            age_range=(self.age_min, self.age_max),
            client_types={toloka.filter.ClientType.ClientType(t) for t in self.client_types},
        )


@message
@dataclass
class WorkerFilter(BaseWorkerFilter, ProtobufSerializer[worker.WorkerFilter]):
    filters: List[WorkerFilterParams] = field(100, default_factory=list)

    @staticmethod
    def serialize(obj: worker.WorkerFilter) -> 'WorkerFilter':
        base_filter = BaseWorkerFilter.serialize(obj)
        return WorkerFilter(
            training_score=base_filter.training_score, filters=[WorkerFilterParams.serialize(f) for f in obj.filters]
        )

    def deserialize(self) -> worker.WorkerFilter:
        base_filter = super(WorkerFilter, self).deserialize()
        return worker.WorkerFilter(
            training_score=base_filter.training_score,
            filters=[f.deserialize() for f in self.filters],
        )


@message
@dataclass
class CustomWorkerFilter(BaseWorkerFilter, ProtobufSerializer[worker.CustomWorkerFilter]):
    filter: Optional[TolokaFilterCondition] = field(100, default=None)

    @staticmethod
    def serialize(obj: worker.CustomWorkerFilter) -> 'CustomWorkerFilter':
        base_filter = BaseWorkerFilter.serialize(obj)
        return CustomWorkerFilter(
            training_score=base_filter.training_score,
            filter=TolokaFilterCondition.serialize(obj.filter) if obj.filter else None,
        )

    def deserialize(self) -> worker.CustomWorkerFilter:
        base_filter = super(CustomWorkerFilter, self).deserialize()
        return worker.CustomWorkerFilter(
            training_score=base_filter.training_score,
            filter=TolokaFilterCondition.deserialize(self.filter) if self.filter else None,
        )


@message
@dataclass
class HumanFilter(ProtobufSerializer[worker.HumanFilter]):
    filter: OneOf_ = one_of(
        expert=part(ExpertFilter, 1),
        worker=part(WorkerFilter, 2),
        custom_worker=part(CustomWorkerFilter, 3),
    )

    @staticmethod
    def serialize(obj: worker.HumanFilter) -> 'HumanFilter':
        f = HumanFilter()
        if isinstance(obj, worker.ExpertFilter):
            f.filter.expert = ExpertFilter.serialize(obj)
        elif isinstance(obj, worker.WorkerFilter):
            f.filter.worker = WorkerFilter.serialize(obj)
        elif isinstance(obj, worker.CustomWorkerFilter):
            f.filter.custom_worker = CustomWorkerFilter.serialize(obj)
        else:
            raise ValueError(f'unexpected filter type: {type(obj)}')
        return f

    def deserialize(self) -> worker.HumanFilter:
        return deserialize_one_of_field(self.filter)


@message
@dataclass
class StaticPricingStrategy(ProtobufSerializer[pricing.StaticPricingStrategy]):
    @staticmethod
    def serialize(obj: pricing.StaticPricingStrategy) -> 'StaticPricingStrategy':
        return StaticPricingStrategy()

    def deserialize(self) -> pricing.StaticPricingStrategy:
        return pricing.StaticPricingStrategy()


@message
@dataclass
class DynamicPricingStrategy(ProtobufSerializer[pricing.DynamicPricingStrategy]):
    max_ratio: float = field(1)

    @staticmethod
    def serialize(obj: pricing.DynamicPricingStrategy) -> 'DynamicPricingStrategy':
        return DynamicPricingStrategy(obj.max_ratio)

    def deserialize(self) -> pricing.DynamicPricingStrategy:
        return pricing.DynamicPricingStrategy(self.max_ratio)


@message
@dataclass
class PricingStrategy(ProtobufSerializer[pricing.PricingStrategy]):
    strategy: OneOf_ = one_of(
        static=part(StaticPricingStrategy, 1),
        dynamic=part(DynamicPricingStrategy, 2),
    )

    @staticmethod
    def serialize(obj: pricing.PricingStrategy) -> 'PricingStrategy':
        strategy = PricingStrategy()
        if isinstance(obj, pricing.StaticPricingStrategy):
            strategy.strategy.static = StaticPricingStrategy.serialize(obj)
        elif isinstance(obj, pricing.DynamicPricingStrategy):
            strategy.strategy.dynamic = DynamicPricingStrategy.serialize(obj)
        else:
            raise ValueError(f'unexpected pricing strategy type: {type(obj)}')
        return strategy

    def deserialize(self) -> pricing.PricingStrategy:
        return deserialize_one_of_field(self.strategy)


@message
@dataclass
class Human(ProtobufSerializer[worker.Human]):
    user_id: str = field(1)
    assignment_id: str = field(2)

    @staticmethod
    def serialize(obj: worker.Human) -> 'Human':
        return Human(user_id=obj.user_id, assignment_id=obj.assignment_id)

    def deserialize(self) -> worker.Human:
        fake_assignment = toloka.Assignment(id=self.assignment_id, user_id=self.user_id)  # TODO: ugly hack
        return worker.Human(fake_assignment)


@message
@dataclass
class Model(ProtobufSerializer[worker.Model]):
    name: str = field(1)
    func: str = field(2)

    @staticmethod
    def serialize(obj: worker.Model) -> 'Model':
        return Model(name=obj.name, func=func_registry[obj.func])

    def deserialize(self) -> worker.Model:
        return worker.Model(name=self.name, func=load_func(self.func))


@message
@dataclass
class Worker(ProtobufSerializer[worker.Worker]):
    worker: OneOf_ = one_of(
        human=part(Human, 1),
        model=part(Model, 2),
    )

    @staticmethod
    def serialize(obj: T) -> 'Worker':
        w = Worker()
        if isinstance(obj, worker.Human):
            w.worker.human = Human.serialize(obj)
        elif isinstance(obj, worker.Model):
            w.worker.model = Model.serialize(obj)
        else:
            raise ValueError(f'unexpected worker type: {type(obj)}')
        return w

    def deserialize(self) -> T:
        return deserialize_one_of_field(self.worker)


@message
@dataclass
class LabelProba(ProtobufSerializer[classification.LabelProba]):
    label: Label = field(1)
    proba: float = field(2)

    @staticmethod
    def serialize(obj: classification.LabelProba) -> 'LabelProba':
        label, proba = obj
        return LabelProba(label=Label.serialize(label), proba=proba)

    def deserialize(self) -> classification.LabelProba:
        return self.label.deserialize(), self.proba


@message
@dataclass
class TaskLabelsProbas(ProtobufSerializer[classification.TaskLabelsProbas]):
    items: List[LabelProba] = field(1, default_factory=list)

    @staticmethod
    def serialize(obj: classification.TaskLabelsProbas) -> 'TaskLabelsProbas':
        return TaskLabelsProbas(items=[LabelProba.serialize((label, proba)) for label, proba in obj.items()])

    def deserialize(self) -> classification.TaskLabelsProbas:
        return {item.label.deserialize(): item.proba for item in self.items}


@message
@dataclass
class StaticOverlap(ProtobufSerializer[classification_loop.StaticOverlap]):
    overlap: int = field(1)

    @staticmethod
    def serialize(obj: classification_loop.StaticOverlap) -> 'StaticOverlap':
        return StaticOverlap(overlap=obj.overlap)

    def deserialize(self) -> classification_loop.StaticOverlap:
        return classification_loop.StaticOverlap(overlap=self.overlap)


@message
@dataclass
class DynamicOverlap(ProtobufSerializer[classification_loop.DynamicOverlap]):
    min_overlap: int = field(1)
    max_overlap: int = field(2)
    confidence: OneOf_ = one_of(
        unified=part(float, 3),
        per_label=part(TaskLabelsProbas, 4),
    )

    @staticmethod
    def serialize(obj: classification_loop.DynamicOverlap) -> 'DynamicOverlap':
        overlap = DynamicOverlap(min_overlap=obj.min_overlap, max_overlap=obj.max_overlap)
        if isinstance(obj.confidence, float):
            overlap.confidence.unified = obj.confidence
        elif isinstance(obj.confidence, dict):
            overlap.confidence.per_label = TaskLabelsProbas.serialize(obj.confidence)
        else:
            raise ValueError(f'unexpected confidence type: {type(obj.confidence)}')
        return overlap

    def deserialize(self) -> classification_loop.DynamicOverlap:
        return classification_loop.DynamicOverlap(
            min_overlap=self.min_overlap,
            max_overlap=self.max_overlap,
            confidence=deserialize_one_of_field(self.confidence),
        )


@message
@dataclass
class Overlap(ProtobufSerializer[classification_loop.Overlap]):
    overlap: OneOf_ = one_of(
        static=part(StaticOverlap, 1),
        dynamic=part(DynamicOverlap, 2),
    )

    @staticmethod
    def serialize(obj: classification_loop.Overlap) -> 'Overlap':
        overlap = Overlap()
        if isinstance(obj, classification_loop.StaticOverlap):
            overlap.overlap.static = StaticOverlap.serialize(obj)
        elif isinstance(obj, classification_loop.DynamicOverlap):
            overlap.overlap.dynamic = DynamicOverlap.serialize(obj)
        else:
            raise ValueError(f'unexpected overlap type: {type(obj)}')
        return overlap

    def deserialize(self) -> classification_loop.Overlap:
        return deserialize_one_of_field(self.overlap)


@message
class AggregationAlgorithm(StrEnum[classification.AggregationAlgorithm]):
    @staticmethod
    def enum_cls() -> Type[classification.AggregationAlgorithm]:
        return classification.AggregationAlgorithm


@message
@dataclass
class ExpertParams(ProtobufSerializer[client.ExpertParams]):
    task_duration_hint: datetime.timedelta = field(1)
    assignment_price: float = field(2)
    real_tasks_count: int = field(3)

    @staticmethod
    def serialize(obj: client.ExpertParams) -> 'ExpertParams':
        return ExpertParams(
            task_duration_hint=obj.task_duration_hint,
            assignment_price=obj.assignment_price,
            real_tasks_count=obj.real_tasks_count,
        )

    def deserialize(self) -> client.ExpertParams:
        return client.ExpertParams(
            task_duration_hint=self.task_duration_hint,
            pricing_config=pricing.PoolPricingConfig(
                assignment_price=self.assignment_price,
                real_tasks_count=self.real_tasks_count,
                control_tasks_count=0,
            ),
        )


@message
@dataclass
class Params(ExpertParams, ProtobufSerializer[client.Params]):
    worker_filter: HumanFilter = field(100)
    control_tasks_count: int = field(101)
    pricing_strategy: PricingStrategy = field(102)
    control: Control = field(103)
    overlap: Overlap = field(104)
    aggregation_algorithm: Optional[AggregationAlgorithm] = field(105, default=None)
    model: Optional[Model] = field(106, default=None)
    task_duration_function: Optional[str] = field(107, default=None)

    @staticmethod
    def serialize(obj: client.Params) -> 'Params':
        expert_params = ExpertParams.serialize(obj)
        return Params(
            task_duration_hint=expert_params.task_duration_hint,
            assignment_price=expert_params.assignment_price,
            real_tasks_count=expert_params.real_tasks_count,
            worker_filter=HumanFilter.serialize(obj.worker_filter),
            control_tasks_count=obj.control_tasks_count,
            pricing_strategy=PricingStrategy.serialize(obj.pricing_strategy),
            control=Control.serialize(obj.control),
            overlap=Overlap.serialize(obj.overlap),
            aggregation_algorithm=AggregationAlgorithm.serialize(obj.aggregation_algorithm)
            if obj.aggregation_algorithm
            else None,
            model=Model.serialize(obj.model) if obj.model else None,
            task_duration_function=func_registry[obj.task_duration_function] if obj.task_duration_function else None,
        )

    def deserialize(self) -> client.Params:
        expert_params = super(Params, self).deserialize()
        return client.Params(
            task_duration_hint=expert_params.task_duration_hint,
            pricing_config=pricing.PoolPricingConfig(
                assignment_price=expert_params.pricing_config.assignment_price,
                real_tasks_count=expert_params.pricing_config.real_tasks_count,
                control_tasks_count=self.control_tasks_count,
            ),
            overlap=Overlap.deserialize(self.overlap),
            control=Control.deserialize(self.control),
            worker_filter=HumanFilter.deserialize(self.worker_filter),
            aggregation_algorithm=AggregationAlgorithm.deserialize(self.aggregation_algorithm)
            if self.aggregation_algorithm
            else None,
            pricing_strategy=PricingStrategy.deserialize(self.pricing_strategy),
            task_duration_function=func_registry_reversed[self.task_duration_function]
            if self.task_duration_function
            else None,
        )


@message
@dataclass
class AssignmentCheckSample(ProtobufSerializer[evaluation.AssignmentCheckSample]):
    max_tasks_to_check: Optional[int] = field(1, default=None)
    assignment_accuracy_finalization_threshold: Optional[float] = field(2, default=None)

    @staticmethod
    def serialize(obj: evaluation.AssignmentCheckSample) -> 'AssignmentCheckSample':
        return AssignmentCheckSample(
            max_tasks_to_check=obj.max_tasks_to_check,
            assignment_accuracy_finalization_threshold=obj.assignment_accuracy_finalization_threshold,
        )

    def deserialize(self) -> evaluation.AssignmentCheckSample:
        return evaluation.AssignmentCheckSample(
            max_tasks_to_check=self.max_tasks_to_check,
            assignment_accuracy_finalization_threshold=self.assignment_accuracy_finalization_threshold,
        )


@message
@dataclass
class AnnotationParams(Params, ProtobufSerializer[client.AnnotationParams]):
    assignment_check_sample: Optional[AssignmentCheckSample] = field(200, default=None)

    @staticmethod
    def serialize(obj: client.AnnotationParams) -> 'AnnotationParams':
        params = Params.serialize(obj)
        return AnnotationParams(
            task_duration_hint=params.task_duration_hint,
            assignment_price=params.assignment_price,
            real_tasks_count=params.real_tasks_count,
            worker_filter=params.worker_filter,
            control_tasks_count=params.control_tasks_count,
            pricing_strategy=params.pricing_strategy,
            control=params.control,
            overlap=params.overlap,
            aggregation_algorithm=params.aggregation_algorithm,
            model=params.model,
            task_duration_function=params.task_duration_function,
            assignment_check_sample=AssignmentCheckSample.serialize(obj.assignment_check_sample)
            if obj.assignment_check_sample
            else None,
        )

    def deserialize(self) -> client.AnnotationParams:
        params = super(AnnotationParams, self).deserialize()
        annotation_params = client.AnnotationParams(
            task_duration_hint=params.task_duration_hint,
            pricing_config=params.pricing_config,
            overlap=params.overlap,
            control=params.control,
            worker_filter=params.worker_filter,
            aggregation_algorithm=params.aggregation_algorithm,
            pricing_strategy=params.pricing_strategy,
            task_duration_function=params.task_duration_function,
            assignment_check_sample=self.assignment_check_sample.deserialize()
            if self.assignment_check_sample
            else None,
        )
        if self.model:
            annotation_params.model = Model.deserialize(self.model)
        return annotation_params



import contextlib
from decimal import Decimal
import io
import itertools
import json
import logging
import os
from typing import List, Union, Type, Dict, Any, Tuple

import pandas as pd

from .. import base, mapping

logger = logging.getLogger(__name__)


@contextlib.contextmanager
def maybe_open(path_or_file: Union[str, os.PathLike, io.TextIOBase]) -> io.TextIOBase:
    if isinstance(path_or_file, str) or isinstance(path_or_file, os.PathLike):
        f = file_to_close = open(path_or_file, 'r')
    else:
        f = path_or_file
        file_to_close = None

    try:
        yield f
    finally:
        if file_to_close:
            file_to_close.close()


def read_tasks(
    path_or_file_or_data: Union[str, os.PathLike, io.TextIOBase, List[Dict[str, Any]]],
    task_mapping: mapping.TaskMapping,
    format: str = 'json',
    has_solutions: bool = False,
) -> List[Union[mapping.Objects, mapping.TaskSingleSolution]]:
    assert format == 'json', 'only JSON format is supported'
    if isinstance(path_or_file_or_data, list):
        rows = path_or_file_or_data
    else:
        with maybe_open(path_or_file_or_data) as f:
            rows = json.load(f, parse_float=Decimal)
    assert isinstance(rows, list)
    result = parse_rows(rows, task_mapping, has_solutions)
    assert_tasks_are_unique(result)
    return result


def get_all_fields(task_mapping: mapping.TaskMapping, has_solutions: bool) -> List[Tuple[str, type]]:
    mappings = task_mapping.input_mapping
    if has_solutions:
        mappings = task_mapping.input_mapping + task_mapping.output_mapping

    def get_field_type(obj_mapping: mapping.ObjectMapping, field: str) -> Type:
        if issubclass(obj_mapping.obj_meta.type, base.Class):
            return str
        return obj_mapping.obj_meta.type.__annotations__.get(field)

    return list(
        itertools.chain(
            *[
                [
                    (task_field, get_field_type(obj_mapping, obj_field))
                    for obj_field, task_field in obj_mapping.obj_task_fields
                ]
                for obj_mapping in mappings
            ]
        )
    )


def parse_rows(
    rows: List[Dict[str, Any]],
    task_mapping: mapping.TaskMapping,
    has_solutions: bool,
) -> List[Union[mapping.Objects, mapping.TaskSingleSolution]]:
    result = []
    for i, row in enumerate(rows):
        try:
            input_objects = task_mapping.from_task_values(row)
            if has_solutions:
                output_objects = task_mapping.from_solution_values(row)
                result.append((input_objects, output_objects))
            else:
                result.append(input_objects)
        except (KeyError, AssertionError, ValueError):
            required_fields = get_all_fields(task_mapping, has_solutions)
            raise ValueError(f'error in file entry #{i + 1}: it must have the following fields: {required_fields}')
    return result


def file_format(task_mapping: mapping.TaskMapping, has_solutions: bool = False) -> pd.DataFrame:
    return pd.DataFrame(
        [{'name': name, 'type': obj_type.__name__} for name, obj_type in get_all_fields(task_mapping, has_solutions)]
    )


def assert_tasks_are_unique(tasks: List[Union[mapping.Objects, mapping.TaskSingleSolution]]):
    assert len(tasks) == len(set(tasks)), 'you have duplicate tasks'



from dataclasses import dataclass, field
import logging
from typing import List, Callable, Dict

import toloka.client as toloka

from ..mapping import Objects, TaskMapping, TaskID
from .common import Worker

logger = logging.getLogger(__name__)


@dataclass
class Model(Worker):
    name: str
    func: Callable[[List[Objects]], List[Objects]]

    @property
    def id(self) -> str:
        return self.name


@dataclass
class ModelWorkspace:
    model: Model
    task_mapping: TaskMapping
    solutions_cache: Dict[TaskID, Objects] = field(init=False, default_factory=dict)

    def get_solutions(
        self,
        pool_input_objects: List[Objects],
        status: toloka.Assignment.Status = toloka.Assignment.SUBMITTED,
    ) -> toloka.Assignment:
        logger.debug(f'model "{self.model.id}" started solving {len(pool_input_objects)} tasks')

        # TODO(DATAFORGE-75)
        # we have no persistent storage for ML worker answers, we have in-memory cache, in case of labeling restart
        # we will re-answer ML worker in same tasks, so we count on model func determinism

        task_id = self.task_mapping.task_id
        not_cached_input_objects = [
            input_objects for input_objects in pool_input_objects if task_id(input_objects) not in self.solutions_cache
        ]
        not_cached_output_objects = self._apply(not_cached_input_objects)

        self.solutions_cache.update(
            {
                task_id(input_objects): output_objects
                for input_objects, output_objects in zip(not_cached_input_objects, not_cached_output_objects)
            }
        )

        pool_output_objects = [self.solutions_cache[task_id(input_objects)] for input_objects in pool_input_objects]

        logger.debug(
            f'model completed tasks solution '
            f'({len(pool_input_objects) - len(not_cached_input_objects)}/{len(pool_input_objects)} cached solutions)'
        )

        return toloka.Assignment(
            tasks=[self.task_mapping.to_task(input_objects) for input_objects in pool_input_objects],
            solutions=[self.task_mapping.to_solution(output_objects) for output_objects in pool_output_objects],
            user_id=self.model.id,
            status=status,
        )

    def _apply(self, pool_input_objects: List[Objects]) -> List[Objects]:
        pool_output_objects = self.model.func(pool_input_objects)

        err_msg_prefix = 'invalid model func:'
        assert len(pool_input_objects) == len(
            pool_output_objects
        ), f'{err_msg_prefix} output list length is not equal to input'
        for output_objects in pool_output_objects:
            assert isinstance(output_objects, tuple), f'{err_msg_prefix} output objects are not tuple'
            assert len(output_objects) == len(
                self.task_mapping.output_mapping
            ), f'{err_msg_prefix} output objects count do not match task function'
            for obj, mapping in zip(output_objects, self.task_mapping.output_mapping):
                assert (
                    type(obj) == mapping.obj_type
                ), f'{err_msg_prefix} object type mismatch, expected {mapping.obj_type}, actual {type(obj)}'

        return pool_output_objects



from datetime import timedelta
import logging
import threading
from typing import List, Optional, Tuple, Dict, Union, Type

# from lzy.api.v1 import LzyRemoteEnv
import toloka.client as toloka

from .. import (
    base,
    classification,
    classification_loop,
    datasource,
    experts,
    feedback_loop,
    mapping,
    mos,
    metrics,
    pool as pool_config,
    project,
    worker,
)
from ..utils import LaunchInterrupter
from .display import get_pool_price, get_annotation_price

# from ..lzy import Whiteboard, add_input_objects, create_pool, run_loop, get_results

from .params import ExpertParams, Params, AnnotationParams
from .task import PreparedTaskSpec, AnnotationTaskSpec, check_project
from .training import find_training_requirement
from .utils import ask, validate_objects_volume
from .relaunch import get_classification_pool, get_annotation_pools

logger = logging.getLogger(__name__)


def create_toloka_client(
    token: str,
    environment: toloka.TolokaClient.Environment = toloka.TolokaClient.Environment.PRODUCTION,
) -> toloka.TolokaClient:
    return toloka.TolokaClient(token=token, environment=environment, retries=20, timeout=(60.0, 240.0))


def launch(
    task_spec: PreparedTaskSpec,
    params: Params,
    input_objects: List[mapping.Objects],
    control_objects: List[mapping.TaskSingleSolution],  # assisted by us
    client: toloka.TolokaClient,
    interactive: bool = False,
    interrupted: Optional[threading.Event] = None,
    name: Optional[str] = None,
) -> Tuple[
    List[Tuple[Optional[classification.TaskLabelsProbas], List[classification.WorkerLabel]]],
    Optional[Dict[str, float]],
]:
    result = _launch(
        task_spec=task_spec,
        params=params,
        input_objects=input_objects,
        control_objects=control_objects,
        client=client,
        interactive=interactive,
        interrupted=interrupted,
        name=name,
    )
    if result is None:
        return [], None

    loop, pool_id = result
    return loop.get_results(pool_id, input_objects)


def launch_mos(
    task_spec: PreparedTaskSpec,
    params: Params,
    input_objects: List[mapping.Objects],
    client: toloka.TolokaClient,
    interactive: bool = False,
    interrupted: Optional[threading.Event] = None,
    inputs_to_metadata: Optional[Dict[mapping.Objects, mos.ObjectsMetadata]] = None,
    name: Optional[str] = None,
) -> Tuple[
    List[Tuple[Optional[classification.TaskLabelsProbas], List[classification.WorkerLabel]]],
    Optional[Dict[str, float]],
    Dict,
]:
    result = _launch(
        task_spec=task_spec,
        params=params,
        input_objects=input_objects,
        control_objects=[],
        client=client,
        interactive=interactive,
        interrupted=interrupted,
        name=name,
        loop_cls=classification_loop.MOSLoop,
        inputs_to_metadata=inputs_to_metadata,
    )

    if result is None:
        return [], None, {}

    loop, pool_id = result

    return loop.get_results(pool_id, input_objects) + (
        loop.assignment_evaluation_strategy.final_assignment_set.get_algorithm_ci(),
    )


def _launch(
    task_spec: PreparedTaskSpec,
    params: Params,
    input_objects: List[mapping.Objects],
    control_objects: List[mapping.TaskSingleSolution],
    client: toloka.TolokaClient,
    interactive: bool = False,
    interrupted: Optional[threading.Event] = None,
    name: Optional[str] = None,
    loop_cls: Type[classification_loop.ClassificationLoop] = classification_loop.ClassificationLoop,
    **kwargs,
) -> Optional[Tuple[classification_loop.ClassificationLoop, str]]:
    assert task_spec.scenario == project.Scenario.DEFAULT, 'You should use this function for crowd markup only'
    assert isinstance(params.task_duration_hint, timedelta)
    assert mapping.validation_enabled
    datasource.assert_tasks_are_unique(input_objects)
    datasource.assert_tasks_are_unique(control_objects)

    prj = check_project(task_spec, client)

    if not validate_objects_volume(
        input_objects,
        control_objects,
        params.task_duration_hint,
        params.pricing_config,
        interactive,
    ):
        return None

    if type(task_spec.function) not in (base.ClassificationFunction, base.SbSFunction):
        raise ValueError(f'unsupported task function: {task_spec.function}')

    # todo: check this some other way - assert isinstance(params, ClassificationLaunchParams)

    interrupter = LaunchInterrupter(client, interrupted or threading.Event())

    if interactive:
        pool_price = get_pool_price(len(input_objects), params.pricing_config, params.overlap, display_formula=True)
        if not ask(f'run classification of {len(input_objects)} objects for {pool_price}?'):
            return None
    loop = loop_cls(
        client,
        task_spec.task_mapping,
        params.classification_loop_params,
        task_spec.lang,
        interrupter,
        with_control_tasks=params.pricing_config.control_tasks_count > 0,
        **kwargs,
    )
    real_tasks_count = params.real_tasks_count
    control_tasks_count = params.control_tasks_count
    training_requirement = find_training_requirement(prj.id, task_spec, client, params)
    pool_cfg = pool_config.ClassificationConfig(
        project_id=prj.id,
        private_name=name or 'pool',
        reward_per_assignment=params.assignment_price,
        task_duration_hint=params.task_duration_hint,
        real_tasks_count=real_tasks_count,
        control_tasks_count=control_tasks_count,
        overlap=params.overlap.min_overlap,
        control_params=params.control,
        worker_filter=params.worker_filter,
        training_requirement=training_requirement,
    )

    pool = get_classification_pool(client, name, loop, pool_cfg, control_objects)
    interrupter.add_pool_id(pool.id)

    plotter = None
    if interactive:
        plotter = metrics.ClassificationMetricsPlotter(
            toloka_client=client,
            stop_event=interrupter.interrupted,
            task_mapping=task_spec.task_mapping,
            pool_id=pool.id,
            task_duration_hint=params.task_duration_hint,
            params=params.classification_loop_params,
            assignment_evaluation_strategy=loop.assignment_evaluation_strategy,
        )

    try:
        loop.add_input_objects(pool.id, input_objects)
        logger.info('classification has started')
        loop.loop(pool.id)
    except KeyboardInterrupt:
        interrupter.shutdown()
        return None
    finally:
        interrupter.interrupted.set()
        if plotter:
            plotter.join()

    return loop, pool.id

    # wb = Whiteboard(
    #     task_spec=task_spec.task_spec,
    #     params=params,
    #     project=prj,
    #     lang=task_spec.lang,
    #     input_objects=input_objects,
    #     control_objects=control_objects)
    #
    # env = LzyRemoteEnv()
    # with env.workflow('classification', whiteboard=wb):
    #     wb.pool = create_pool(loop, control_objects, pool_cfg)
    #     add_input_objects(loop, input_objects, wb.pool)
    #     run_loop(loop, wb.pool)
    #     wb.results = get_results(loop, input_objects, wb.pool)
    #     wb_id = wb.__id__
    #
    # return wb_id


# tmp routine, while we didn't implement pipeline in lzy
def select_control_tasks(
    input_objects: List[mapping.Objects],
    results: Union[
        List[Tuple[classification.TaskLabelsProbas, List[classification.WorkerLabel]]],
        List[List[feedback_loop.Solution]],
    ],
    min_confidence: float,
    min_overlap: int = 1,
) -> List[mapping.TaskSingleSolution]:
    control_tasks = []
    if not isinstance(results[0][0], feedback_loop.Solution):
        for task_input_objects, (labels_probas, raw_labels) in zip(input_objects, results):
            label, proba = classification.get_most_probable_label(labels_probas)
            if len(raw_labels) >= min_overlap and proba >= min_confidence:
                control_tasks.append((task_input_objects, (label,)))
        return control_tasks

    for result in results:
        if not result.solutions:
            continue
        task, best_solution = result.task, result.solutions[0]
        # if we have non-trivial check_sample, there can be good annotations without evaluation - unknown confidence
        # however, in case of min_confidence == 0 we may want to receive TaskSingleSolution for _each_ input object
        if not best_solution.evaluation:
            continue
        solution, proba = best_solution.solution, best_solution.evaluation.confidence
        raw_labels = best_solution.evaluation.worker_labels
        if len(raw_labels) >= min_overlap and proba >= min_confidence:
            control_tasks.append((task, solution))
    return control_tasks


# todo реальные пулы нельзя создавать с 0 стоимостью. мб можно все делать в сэндбоксе
# todo вообще нужно корректировать task_duration с учетом сценариев
def launch_experts(
    task_spec: PreparedTaskSpec,
    params: ExpertParams,
    input_objects: Union[List[mapping.Objects], List[mapping.TaskSingleSolution]],
    case: experts.ExpertCase,
    client: toloka.TolokaClient,
    name: Optional[str] = None,
    interactive: bool = False,
) -> List[Tuple[Union[mapping.TaskSingleSolution, mapping.Objects], mapping.TaskMultipleSolutions]]:
    assert task_spec.scenario != project.Scenario.DEFAULT, 'You should use this function for expert markup only'
    assert isinstance(params.task_duration_hint, timedelta)
    assert mapping.validation_enabled

    prj = check_project(task_spec, client)

    assignment_price = params.assignment_price
    pipeline = experts.ExpertPipeline(
        client, task_spec.function, task_spec.task_mapping, task_spec.lang, scenario=task_spec.scenario
    )

    skills = experts.get_skills(task_spec.id, task_spec.lang, client)
    assert skills, f'No suitable skills found for task spec {task_spec.id} and lang {task_spec.lang}'
    logger.debug(f'Using found skills: {skills}')

    if interactive:
        pool_price = get_pool_price(
            len(input_objects),
            params.pricing_config,
            classification_loop.StaticOverlap(overlap=1),
            display_formula=True,
        )
        if not ask(f'run expert labeling of {len(input_objects)} objects for {pool_price}?'):
            return []

    interrupter = LaunchInterrupter(client, threading.Event())

    pool_cfg = pool_config.ExpertConfig(
        project_id=prj.id,
        private_name=name or 'pool',
        public_description=case.label[task_spec.lang],
        reward_per_assignment=assignment_price,
        task_duration_hint=params.task_duration_hint,
        real_tasks_count=params.pricing_config.get_tasks_count(),
        worker_filter=worker.ExpertFilter([skill for (_, skill) in skills]),
    )

    pool = get_classification_pool(client, name, pipeline, pool_cfg)
    interrupter.add_pool_id(pool.id)

    # todo some extra functionality will be lost here without lzy graph rearrangements - e.q. pre-/post-swaps in sbs
    try:
        pipeline.add_input_objects(pool.id, input_objects)
        logger.info('expert labeling has started')
        pipeline.loop(pool.id)
    except KeyboardInterrupt:
        interrupter.shutdown(reraise=False)

    return pipeline.get_results(pool.id, input_objects)


def launch_annotation(
    task_spec: AnnotationTaskSpec,
    params: AnnotationParams,
    check_params: Params,
    input_objects: List[mapping.Objects],
    control_objects: List[mapping.TaskSingleSolution],  # assisted by us
    client: toloka.TolokaClient,
    interactive: bool = False,
    interrupted: Optional[threading.Event] = None,
    name: Optional[str] = None,
    s3: Optional[datasource.S3] = None,
) -> Tuple[List[List[feedback_loop.Solution]], Optional[Dict[str, float]]]:
    assert task_spec.scenario == project.Scenario.DEFAULT, 'You should use this function for crowd markup only'
    assert isinstance(params.task_duration_hint, timedelta)
    assert isinstance(check_params.task_duration_hint, timedelta)
    assert mapping.validation_enabled
    datasource.assert_tasks_are_unique(input_objects)
    datasource.assert_tasks_are_unique(control_objects)

    markup_prj = check_project(task_spec, client)
    check_prj = check_project(task_spec.check, client)
    # todo: validate_objects_volume

    for obj in task_spec.function.get_outputs():
        if obj.type.is_media():
            assert s3 is not None, 'S3 client is required for media-output tasks'

    interrupter = LaunchInterrupter(client, interrupted or threading.Event())

    if interactive:
        annotation_price = get_annotation_price(
            input_objects_count=len(input_objects),
            markup_config=params.pricing_config,
            check_config=check_params.pricing_config,
            markup_overlap=params.overlap,
            check_overlap=check_params.overlap,
            assignment_check_sample=params.assignment_check_sample,
            display_formula=True,
        )
        if not ask(f'run annotation of {len(input_objects)} objects for {annotation_price}?'):
            return [], None

    fb_loop = feedback_loop.FeedbackLoop(
        pool_input_objects=input_objects,
        markup_task_mapping=task_spec.task_mapping,
        check_task_mapping=task_spec.check.task_mapping,
        markup_params=params.feedback_loop_params,
        check_params=check_params.classification_loop_params,
        client=client,
        lang=task_spec.lang,
        interrupter=interrupter,
        s3=s3,
        model_markup=params.model,
        model_check=check_params.model,
    )

    check_training_requirement = find_training_requirement(check_prj.id, task_spec.check, client, check_params)
    markup_training_requirement = find_training_requirement(markup_prj.id, task_spec, client, params)

    check_pool_cfg = pool_config.ClassificationConfig(
        project_id=check_prj.id,
        private_name=name or 'check pool',
        reward_per_assignment=check_params.assignment_price,
        task_duration_hint=check_params.task_duration_hint,
        real_tasks_count=check_params.real_tasks_count,
        control_tasks_count=check_params.control_tasks_count,
        overlap=check_params.overlap.min_overlap,
        control_params=check_params.control,
        worker_filter=check_params.worker_filter,
        training_requirement=check_training_requirement,
    )

    markup_pool_cfg = pool_config.MarkupConfig(
        project_id=markup_prj.id,
        private_name=name or 'markup pool',
        reward_per_assignment=params.assignment_price,
        task_duration_hint=params.task_duration_hint,
        real_tasks_count=params.real_tasks_count,
        worker_filter=params.worker_filter,
        training_requirement=markup_training_requirement,
        control_params=params.control,
        first_attempt_by_model_worker=params.model is not None,
    )
    markup_pool, check_pool = get_annotation_pools(
        client, name, fb_loop, check_pool_cfg, markup_pool_cfg, control_objects
    )
    interrupter.add_pool_id(markup_pool.id)
    interrupter.add_pool_id(check_pool.id)

    plotter = None
    if interactive:
        plotter = metrics.AnnotationMetricsPlotter(
            toloka_client=client,
            stop_event=interrupter.interrupted,
            task_spec=task_spec,
            check_pool_id=check_pool.id,
            markup_pool_id=markup_pool.id,
            check_task_duration_hint=check_params.task_duration_hint,
            markup_task_duration_hint=params.task_duration_hint,
            evaluation=fb_loop.evaluation,
        )

    try:
        logger.info('annotation has started')
        fb_loop.loop(markup_pool.id, check_pool.id)
    except KeyboardInterrupt:
        interrupter.shutdown()
        return [], None
    finally:
        interrupter.interrupted.set()
        if plotter:
            plotter.join()

    return fb_loop.get_results(markup_pool.id, check_pool.id)
